from __future__ import annotations

import os
import re
import typing as t
import uuid
from abc import ABC
from datetime import datetime
from datetime import timezone
from functools import lru_cache
from platform import platform
from platform import python_version

import attr
import attr.converters
import psutil
import yaml
from simple_di import Provide
from simple_di import inject

from ...cloud.config import CloudClientConfig
from ...configuration import BENTOML_VERSION
from ...configuration.containers import BentoMLContainer
from ...utils import is_jupyter
from ...utils.cattr import bentoml_cattr

if t.TYPE_CHECKING:
    from pathlib import Path

    P = t.ParamSpec("P")
    GenericFunction = t.Callable[P, t.Any]

# Refer to bentoml/yatai-deployment-operator/common/consts/consts.go
ENV_YATAI_VERSION = "YATAI_T_VERSION"
ENV_YATAI_ORG_UID = "YATAI_T_ORG_UID"
ENV_YATAI_DEPLOYMENT_UID = "YATAI_T_DEPLOYMENT_UID"
ENV_YATAI_CLUSTER_UID = "YATAI_T_CLUSTER_UID"


@lru_cache(maxsize=1)
def get_platform() -> str:
    return platform(aliased=True)


@lru_cache(maxsize=1)
def get_python_version() -> str:
    return python_version()


@attr.define
class ClientInfo:
    id: str
    creation_timestamp: datetime


@inject
@lru_cache(maxsize=1)
def get_client_info(
    bentoml_home: str = Provide[BentoMLContainer.bentoml_home],
) -> t.Optional[ClientInfo]:
    CLIENT_INFO_FILE_PATH = os.path.join(bentoml_home, "client_id")

    if os.path.exists(CLIENT_INFO_FILE_PATH):
        with open(CLIENT_INFO_FILE_PATH, "r", encoding="utf-8") as f:
            client_info = yaml.safe_load(f)
        return bentoml_cattr.structure(client_info, ClientInfo)
    else:
        # Create new client id
        new_client_info = ClientInfo(
            id=str(uuid.uuid4()),
            creation_timestamp=datetime.now(timezone.utc),
        )
        # write client info to ~/bentoml/client_id
        with open(CLIENT_INFO_FILE_PATH, "w", encoding="utf-8") as f:
            yaml.dump(attr.asdict(new_client_info), stream=f)

        return new_client_info


@inject
@lru_cache(maxsize=1)
def get_yatai_user_email(
    cloud_config: Path = Provide[BentoMLContainer.cloud_config],
) -> str | None:
    if cloud_config.exists():
        return CloudClientConfig.get_config(cloud_config).get_context().email


@lru_cache(maxsize=1)
def is_interactive() -> bool:
    import __main__ as main

    return not hasattr(main, "__file__")


@attr.define
class CommonProperties:
    # when the event is triggered
    timestamp: datetime = attr.field(factory=lambda: datetime.now(timezone.utc))

    # environment related
    platform: str = attr.field(factory=get_platform)
    bentoml_version: str = attr.field(default=BENTOML_VERSION)
    python_version: str = attr.field(factory=get_python_version)
    is_interactive: bool = attr.field(factory=is_interactive)
    in_notebook: bool = attr.field(factory=is_jupyter)

    # resource related
    memory_usage_percent: float = attr.field(init=False)
    total_memory_in_mb: float = attr.field(init=False)

    # client related
    client: ClientInfo = attr.field(factory=get_client_info)
    yatai_user_email: t.Optional[str] = attr.field(factory=get_yatai_user_email)
    yatai_version: t.Optional[str] = attr.field(
        default=os.environ.get(ENV_YATAI_VERSION, None)
    )
    yatai_org_uid: t.Optional[str] = attr.field(
        default=os.environ.get(ENV_YATAI_ORG_UID, None)
    )
    yatai_cluster_uid: t.Optional[str] = attr.field(
        default=os.environ.get(ENV_YATAI_CLUSTER_UID, None)
    )
    yatai_deployment_uid: t.Optional[str] = attr.field(
        default=os.environ.get(ENV_YATAI_DEPLOYMENT_UID, None)
    )

    def __attrs_post_init__(self):
        self.total_memory_in_mb = int(psutil.virtual_memory().total / 1024.0 / 1024.0)
        proc = psutil.Process(os.getpid())
        with proc.oneshot():
            self.memory_usage_percent = proc.memory_percent()


class EventMeta(ABC):
    @property
    def event_name(self):
        # camel case to snake case
        event_name = re.sub(r"(?<!^)(?=[A-Z])", "_", self.__class__.__name__).lower()
        # remove "_event" suffix
        suffix_to_remove = "_event"
        if event_name.endswith(suffix_to_remove):
            event_name = event_name[: -len(suffix_to_remove)]
        return event_name


@attr.define
class CliEvent(EventMeta):
    cmd_group: str
    cmd_name: str
    duration_in_ms: float = attr.field(default=0)
    error_type: t.Optional[str] = attr.field(default=None)
    return_code: t.Optional[int] = attr.field(default=None)


@attr.define
class BentoBuildEvent(CliEvent):
    bento_creation_timestamp: t.Optional[datetime] = attr.field(default=None)
    bento_size_in_kb: float = attr.field(default=0)
    model_size_in_kb: float = attr.field(default=0)

    num_of_models: int = attr.field(default=0)
    num_of_runners: int = attr.field(default=0)
    model_types: t.List[str] = attr.field(factory=list)
    runnable_types: t.List[str] = attr.field(factory=list)


@attr.define
class ModelSaveEvent(EventMeta):
    module: str
    model_size_in_kb: float


# serve_kind determines different type of serving scenarios
SERVE_KIND = ["grpc", "http"]
# components are different components to be tracked.
COMPONENT_KIND = ["standalone", "api_server", "runner"]


@attr.define
class ServeInitEvent(EventMeta):
    serve_id: str
    production: bool
    serve_from_bento: bool
    serve_from_server_api: bool

    bento_creation_timestamp: t.Optional[datetime]
    serve_kind: str = attr.field(validator=attr.validators.in_(SERVE_KIND))
    num_of_models: int = attr.field(default=0)
    num_of_runners: int = attr.field(default=0)
    num_of_apis: int = attr.field(default=0)
    model_types: t.List[str] = attr.field(factory=list)
    runnable_types: t.List[str] = attr.field(factory=list)
    api_input_types: t.List[str] = attr.field(factory=list)
    api_output_types: t.List[str] = attr.field(factory=list)


@attr.define
class ServeUpdateEvent(EventMeta):
    serve_id: str
    production: bool
    triggered_at: datetime
    duration_in_seconds: int
    serve_kind: str = attr.field(validator=attr.validators.in_(SERVE_KIND))
    serve_from_server_api: bool
    component: str = attr.field(
        validator=attr.validators.and_(
            attr.validators.instance_of(str), attr.validators.in_(COMPONENT_KIND)
        )
    )
    metrics: t.List[t.Any] = attr.field(factory=list)


ALL_EVENT_TYPES = t.Union[
    CliEvent,
    ModelSaveEvent,
    BentoBuildEvent,
    ServeInitEvent,
    ServeUpdateEvent,
    EventMeta,
]


@attr.define
class TrackingPayload:
    session_id: str
    event_properties: ALL_EVENT_TYPES
    common_properties: CommonProperties
    event_type: str

    def to_dict(self):
        return bentoml_cattr.unstructure(self)



from __future__ import annotations

import contextlib
import typing as t
from threading import Thread
from typing import TYPE_CHECKING
from urllib.parse import urlparse

import attrs
from circus.arbiter import Arbiter as _Arbiter

if TYPE_CHECKING:
    from circus.sockets import CircusSocket
    from circus.watcher import Watcher

__all__ = [
    "create_circus_socket_from_uri",
    "create_standalone_arbiter",
]


class Arbiter(_Arbiter):
    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        super().__init__(*args, **kwargs)
        self.exit_stack = contextlib.ExitStack()

    def start(self, cb: t.Callable[[t.Any], t.Any] | None = None) -> None:
        self.exit_stack.__enter__()
        fut = super().start(cb)
        if exc := fut.exception():
            raise exc

    def stop(self) -> None:
        self.exit_stack.__exit__(None, None, None)
        return super().stop()


class ThreadedArbiter(Arbiter, Thread):
    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        Arbiter.__init__(self, *args, **kwargs)
        Thread.__init__(self, daemon=True)
        self.__cb: t.Optional[t.Callable[[t.Any], t.Any]] = None

    def start(self, cb: t.Callable[[t.Any], t.Any] | None = None) -> None:
        self.__cb = cb
        Thread.start(self)

    def run(self) -> None:
        # reset the loop in thread
        self.loop = None
        self.ctrl.loop = self._ensure_ioloop()  # type: ignore[union-attr]
        Arbiter.start(self, self.__cb)

    def stop(self) -> None:
        self.loop.add_callback(Arbiter.stop, self)  # type: ignore[union-attr]
        Thread.join(self)


def create_circus_socket_from_uri(
    uri: str, *args: t.Any, name: str = "", **kwargs: t.Any
) -> CircusSocket:
    from circus.sockets import CircusSocket

    from ..uri import uri_to_path

    parsed = urlparse(uri)
    if parsed.scheme in ("file", "unix"):
        return CircusSocket(
            name=name,
            path=uri_to_path(uri),
            *args,
            **kwargs,
        )
    elif parsed.scheme == "tcp":
        return CircusSocket(
            name=name,
            host=parsed.hostname,
            port=parsed.port,
            *args,
            **kwargs,
        )
    else:
        raise ValueError(f"Unsupported URI scheme: {parsed.scheme}")


def create_standalone_arbiter(
    watchers: list[Watcher], *, threaded: bool = False, **kwargs: t.Any
) -> Arbiter:
    from .. import reserve_free_port

    arbiter_cls = ThreadedArbiter if threaded else Arbiter

    with reserve_free_port() as endpoint_port:
        with reserve_free_port() as pubsub_port:
            return arbiter_cls(
                watchers,
                endpoint=f"tcp://127.0.0.1:{endpoint_port}",
                pubsub_endpoint=f"tcp://127.0.0.1:{pubsub_port}",
                # XXX: Currently ,the default check_delay will always raise ConflictError. This probably has to do with
                # the runners server is not ready in time when the controller run healthcheck.
                check_delay=kwargs.pop("check_delay", 10),
                **kwargs,
            )


@attrs.frozen
class Server:
    url: str
    arbiter: Arbiter = attrs.field(repr=False)

    def start(self) -> None:
        pass

    def stop(self) -> None:
        self.arbiter.stop()

    @property
    def running(self) -> bool:
        return self.arbiter.running

    def __enter__(self) -> Server:
        return self

    def __exit__(self, exc_type: t.Any, exc_value: t.Any, traceback: t.Any) -> None:
        self.stop()



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any

import attrs

from ..tasks import ResultStatus

if TYPE_CHECKING:
    from .http import AsyncHTTPClient
    from .http import ClientEndpoint
    from .http import SyncHTTPClient


@attrs.frozen
class Task:
    id: str
    endpoint: ClientEndpoint = attrs.field(repr=False)
    client: SyncHTTPClient = attrs.field(repr=False)

    def get_status(self) -> ResultStatus:
        return self.client._get_task_status(self.endpoint, self.id)

    def cancel(self) -> None:
        return self.client._cancel_task(self.endpoint, self.id)

    def get(self) -> Any:
        return self.client._get_task_result(self.endpoint, self.id)

    def retry(self) -> Task:
        return self.client._retry_task(self.endpoint, self.id)


@attrs.frozen
class AsyncTask:
    id: str
    endpoint: ClientEndpoint = attrs.field(repr=False)
    client: AsyncHTTPClient = attrs.field(repr=False)

    async def get_status(self) -> ResultStatus:
        return await self.client._get_task_status(self.endpoint, self.id)

    async def cancel(self) -> None:
        return await self.client._cancel_task(self.endpoint, self.id)

    async def get(self) -> Any:
        return await self.client._get_task_result(self.endpoint, self.id)

    async def retry(self) -> AsyncTask:
        return await self.client._retry_task(self.endpoint, self.id)



from __future__ import annotations

import logging
import typing as t
from http import HTTPStatus

import attr
import yaml
from simple_di import Provide
from simple_di import inject

from ...exceptions import CloudRESTApiClientError
from ..configuration.containers import BentoMLContainer
from ..utils.cattr import bentoml_cattr
from .client import RestApiClient

logger = logging.getLogger(__name__)

default_context_name = "default"

if t.TYPE_CHECKING:
    from pathlib import Path


@attr.define
class CloudClientContext:
    name: str
    endpoint: str
    api_token: str
    email: t.Optional[str] = attr.field(default=None)

    def get_rest_api_client(self) -> RestApiClient:
        return RestApiClient(self.endpoint, self.api_token)

    def get_email(self) -> str:
        if not self.email:
            cli = self.get_rest_api_client()
            user = cli.v1.get_current_user()
            if user is None:
                raise CloudRESTApiClientError(
                    "Unable to get current user from yatai server"
                )
            self.email = user.email
            self.save(ignore_warning=True)
        return self.email

    def save(self, *, ignore_warning: bool = False) -> None:
        config = CloudClientConfig.get_config()
        for idx, ctx in enumerate(config.contexts):
            if ctx.name == self.name:
                if not ignore_warning:
                    logger.warning(
                        "Overriding existing cloud context config: %s", ctx.name
                    )
                config.contexts[idx] = self
                break
        else:
            config.contexts.append(self)
        config.to_yaml_file()


DEFAULT_ENDPOINT = "https://cloud.bentoml.com"


@attr.define
class CloudClientConfig:
    contexts: t.List[CloudClientContext] = attr.field(factory=list)
    current_context_name: str = attr.field(default=default_context_name)

    def get_context(self, context: t.Optional[str] = None) -> CloudClientContext:
        from os import environ

        if "BENTO_CLOUD_API_KEY" in environ:
            return CloudClientContext(
                name="__env__",
                endpoint=environ.get("BENTO_CLOUD_API_ENDPOINT", DEFAULT_ENDPOINT),
                api_token=environ["BENTO_CLOUD_API_KEY"],
            )
        if context is None:
            context = self.current_context_name
        for ctx in self.contexts:
            if ctx.name == context:
                return ctx
        raise CloudRESTApiClientError(
            f"No cloud context {context} found",
            error_code=HTTPStatus.UNAUTHORIZED,
        )

    def set_current_context(self, context: str | None) -> CloudClientContext:
        """Set current context to a new context default."""
        try:
            new_context = self.get_context(context)
        except CloudRESTApiClientError as err:
            raise err from None
        new_config = attr.evolve(self, current_context_name=new_context.name)
        new_config.to_yaml_file()
        return new_context

    @inject
    def to_yaml_file(
        self, cloud_config: Path = Provide[BentoMLContainer.cloud_config]
    ) -> None:
        """Converts this config to yaml file at given path object.
        Note that this method will not return the config object.
        """
        with cloud_config.open("w") as f:
            yaml.dump(bentoml_cattr.unstructure(self), stream=f)

    @classmethod
    def default_config(cls) -> CloudClientConfig:
        """Iinitialize a default config and write it to default yaml path."""
        config = cls(contexts=[], current_context_name=default_context_name)
        config.to_yaml_file()
        return config

    @classmethod
    @inject
    def get_config(
        cls, cloud_config: Path = Provide[BentoMLContainer.cloud_config]
    ) -> CloudClientConfig:
        if not cloud_config.exists():
            return cls.default_config()
        with cloud_config.open("r") as stream:
            try:
                yaml_content = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                logger.error(exc)
                raise

            if not yaml_content:
                return cls.default_config()
            return bentoml_cattr.structure(yaml_content, cls)


def get_rest_api_client(context: str | None = None) -> RestApiClient:
    cfg = CloudClientConfig.get_config()
    ctx = cfg.get_context(context)
    return ctx.get_rest_api_client()



from __future__ import annotations

import logging
import os
import typing as t
from types import ModuleType
from typing import TYPE_CHECKING

import attr
import numpy as np

import bentoml
from bentoml import Tag
from bentoml._internal.models.model import ModelContext
from bentoml._internal.utils import deprecated
from bentoml._internal.utils.pkg import get_pkg_version
from bentoml.exceptions import BentoMLException
from bentoml.exceptions import InvalidArgument
from bentoml.exceptions import MissingDependencyException
from bentoml.exceptions import NotFound
from bentoml.models import ModelOptions as BaseModelOptions
from bentoml.models import get as get

if TYPE_CHECKING:
    from typing_extensions import Unpack

    from _bentoml_sdk import Service
    from _bentoml_sdk import ServiceConfig
    from bentoml.types import ModelSignature
    from bentoml.types import ModelSignatureDict


try:
    import catboost as cb
except ImportError:  # pragma: no cover
    raise MissingDependencyException(
        "'catboost' is required in order to use module 'bentoml.catboost', install catboost with 'pip install catboost'. For more information, refer to https://catboost.ai/en/docs/concepts/installation."
    )

MODULE_NAME = "bentoml.catboost"
MODEL_FILENAME = "saved_model.cbm"
DEFAULT_MODEL_TRAINING_CLASS_NAME = "CatBoost"
API_VERSION = "v1"

logger = logging.getLogger(MODULE_NAME)


def load_model(bento_model: str | Tag | bentoml.Model) -> cb.CatBoost:
    """
    Load the CatBoost model with the given tag from the local BentoML model store.

    Args:
        bento_model (``str`` ``|`` :obj:`~bentoml.Tag` ``|`` :obj:`~bentoml.Model`):
            Either the tag of the model to get from the store, or a BentoML `~bentoml.Model`
            instance to load the model from.

    Returns:
        :obj:`~catboost.CatBoost`: The CatBoost model loaded from the model store or BentoML :obj:`~bentoml.Model`.

    Example:

    .. code-block:: python

        import bentoml
        # target model must be from the BentoML model store
        booster = bentoml.catboost.load_model("my_catboost_model")
    """  # noqa: LN001
    if not isinstance(bento_model, bentoml.Model):
        bento_model = bentoml.models.get(bento_model)

    if bento_model.info.module not in (MODULE_NAME, __name__):
        raise NotFound(
            f"Model {bento_model.tag} was saved with module {bento_model.info.module}, not loading with {MODULE_NAME}."
        )

    model_file = bento_model.path_of(MODEL_FILENAME)
    cb_class_name: str = bento_model.info.options.training_class_name  # type: ignore
    cb_class: t.Type[cb.CatBoost] = getattr(cb, cb_class_name)
    if not issubclass(cb_class, cb.CatBoost):
        raise BentoMLException(f"{cb_class_name} is not a valid CatBoost class.")
    cb_instance = cb_class()
    booster: cb.CatBoost = cb_instance.load_model(fname=model_file)
    return booster


@attr.define
class ModelOptions(BaseModelOptions):
    training_class_name: str = attr.field(factory=str)


def save_model(
    name: Tag | str,
    model: cb.CatBoost,
    *,
    signatures: dict[str, ModelSignatureDict] | None = None,
    labels: dict[str, str] | None = None,
    custom_objects: dict[str, t.Any] | None = None,
    external_modules: t.List[ModuleType] | None = None,
    metadata: dict[str, t.Any] | None = None,
) -> bentoml.Model:
    """
    Save an CatBoost model instance to the BentoML model store.

    Args:
        name:
            The name to give to the model in the BentoML store. This must be a valid
            :obj:`~bentoml.Tag` name.
        model:
            The CatBoost model to be saved.
        signatures:
            Signatures of predict methods to be used. If not provided, the signatures default to
            ``{"predict": {"batchable": False}}``. See :obj:`~bentoml.types.ModelSignature` for more
            details.
        labels:
            A default set of management labels to be associated with the model. An example is
            ``{"training-set": "data-1"}``.
        custom_objects:
            Custom objects to be saved with the model. An example is
            ``{"my-normalizer": normalizer}``.

            Custom objects are currently serialized with cloudpickle, but this implementation is
            subject to change.
        external_modules (:code:`List[ModuleType]`, `optional`, default to :code:`None`):
            user-defined additional python modules to be saved alongside the model or custom objects,
            e.g. a tokenizer module, preprocessor module, model configuration module
        metadata:
            Metadata to be associated with the model. An example is ``{"max_depth": 2}``.

            Metadata is intended for display in model management UI and therefore must be a default
            Python type, such as ``str`` or ``int``.

    Returns:
        :obj:`~bentoml.Tag`: A :obj:`tag` with a format `name:version` where `name` is the
        user-defined model's name, and a generated `version` by BentoML.

    Example:

    .. code-block:: python

        import bentoml
        import numpy as np

        from catboost import CatBoostClassifier, Pool

        # initialize data
        train_data = np.random.randint(0, 100, size=(100, 10))

        train_labels = np.random.randint(0, 2, size=(100))

        test_data = catboost_pool = Pool(train_data, train_labels)

        model = CatBoostClassifier(iterations=2,
                                   depth=2,
                                   learning_rate=1,
                                   loss_function='Logloss',
                                   verbose=True)
        # train the model
        model.fit(train_data, train_labels)

        # save the model to the BentoML model store
        bento_model = bentoml.catboost.save_model("my_catboost_model", model)
    """
    if not isinstance(model, cb.CatBoost):
        raise TypeError(f"Given model ({model}) is not a catboost.CatBoost.")

    context: ModelContext = ModelContext(
        framework_name="catboost",
        framework_versions={"catboost": get_pkg_version("catboost")},
    )

    if signatures is None:
        signatures = {
            "predict": {"batchable": False},
        }
        logger.info(
            'Using the default model signature for CatBoost (%s) for model "%s".',
            signatures,
            name,
        )

    options = ModelOptions(
        training_class_name=model.__class__.__name__,
    )

    with bentoml.models._create(  # type: ignore
        name,
        module=MODULE_NAME,
        api_version=API_VERSION,
        signatures=signatures,
        labels=labels,
        custom_objects=custom_objects,
        external_modules=external_modules,
        metadata=metadata,
        context=context,
        options=options,
    ) as bento_model:
        model.save_model(bento_model.path_of(MODEL_FILENAME))  # type: ignore (incomplete CatBoost types)

        return bento_model


@deprecated(suggestion="Use `get_service` instead.")
def get_runnable(bento_model: bentoml.Model) -> t.Type[bentoml.Runnable]:
    """
    Private API: use :obj:`~bentoml.Model.to_runnable` instead.
    """

    class CatBoostRunnable(bentoml.Runnable):
        SUPPORTED_RESOURCES = ("nvidia.com/gpu", "cpu")
        SUPPORTS_CPU_MULTI_THREADING = True

        predict_params: t.Dict[str, t.Any]

        def __init__(self):
            super().__init__()
            self.model = load_model(bento_model)

            self.predict_params = {
                "task_type": "CPU",
            }

            # check for resources
            available_gpus = os.getenv("CUDA_VISIBLE_DEVICES", "")
            if available_gpus not in ("", "-1"):
                self.predict_params["task_type"] = "GPU"
            else:
                nthreads = os.getenv("OMP_NUM_THREADS")
                if nthreads is not None and nthreads != "":
                    nthreads = max(int(nthreads), 1)
                else:
                    nthreads = -1
                self.predict_params["thread_count"] = nthreads

            self.predict_fns: dict[str, t.Callable[..., t.Any]] = {}
            for method_name in bento_model.info.signatures:
                try:
                    self.predict_fns[method_name] = getattr(self.model, method_name)
                except AttributeError:
                    raise InvalidArgument(
                        f"No method with name {method_name} found for CatBoost model of type {self.model.__class__}"
                    )

    def add_runnable_method(method_name: str, options: ModelSignature):
        def _run(self: CatBoostRunnable, input_data: t.Any) -> t.Any:
            if not isinstance(input_data, cb.Pool):
                input_data = cb.Pool(input_data)
            res = self.predict_fns[method_name](input_data, **self.predict_params)
            return np.asarray(res)  # type: ignore (incomplete np types)

        CatBoostRunnable.add_method(
            _run,
            name=method_name,
            batchable=options.batchable,
            batch_dim=options.batch_dim,
            input_spec=options.input_spec,
            output_spec=options.output_spec,
        )

    for method_name, options in bento_model.info.signatures.items():
        add_runnable_method(method_name, options)

    return CatBoostRunnable


def get_service(model_name: str, **config: Unpack[ServiceConfig]) -> Service[t.Any]:
    """
    Get a BentoML service for the catboost model given by name.

    Args:
        model_name (``str``):
            The name of the model to get the service for.
        **config (``Unpack[ServiceConfig]``):
            Configuration options for the service.
    Returns:
        A BentoML service instance that wraps the CatBoost model.
    Example:

    .. code-block:: python

        import bentoml

        service = bentoml.catboost.get_service("my_catboost_model")
    """

    @bentoml.service(**config)
    class CatBoostService:
        bento_model = bentoml.models.get(model_name)

        def __init__(self) -> None:
            self.model = load_model(self.bento_model)
            self.predict_params = {"task_type": "CPU"}

            # check for resources
            available_gpus = os.getenv("CUDA_VISIBLE_DEVICES", "")
            if available_gpus not in ("", "-1"):
                self.predict_params["task_type"] = "GPU"
            else:
                nthreads = os.getenv("OMP_NUM_THREADS")
                if nthreads is not None and nthreads != "":
                    nthreads = max(int(nthreads), 1)
                else:
                    nthreads = -1
                self.predict_params["thread_count"] = nthreads

        @bentoml.api
        def predict(self, input_data: np.ndarray) -> np.ndarray:
            rv = self.model.predict(data=cb.Pool(input_data), **self.predict_params)
            return np.asarray(rv)  # type: ignore (incomplete np types)

    return CatBoostService



from __future__ import annotations

import abc
import typing as t

import attrs
from fs.base import FS
from simple_di import Provide
from simple_di import inject

from bentoml._internal.bento.bento import BentoModelInfo
from bentoml._internal.cloud.client import RestApiClient
from bentoml._internal.cloud.schemas.modelschemas import LabelItemSchema
from bentoml._internal.cloud.schemas.modelschemas import ModelManifestSchema
from bentoml._internal.cloud.schemas.schemasv1 import CreateModelSchema
from bentoml._internal.configuration.containers import BentoMLContainer
from bentoml._internal.models import Model as StoredModel
from bentoml._internal.models import ModelStore
from bentoml._internal.models.model import copy_model
from bentoml._internal.tag import Tag
from bentoml._internal.types import PathType
from bentoml._internal.utils.filesystem import calc_dir_size
from bentoml.exceptions import NotFound

if t.TYPE_CHECKING:
    from bentoml._internal.cloud.schemas.schemasv1 import ModelSchema

T = t.TypeVar("T")


class Model(abc.ABC, t.Generic[T]):
    """A model reference to a artifact in various registries."""

    @abc.abstractmethod
    def to_info(self, alias: str | None = None) -> BentoModelInfo:
        """Return the model info object."""

    @classmethod
    @abc.abstractmethod
    def from_info(cls, info: BentoModelInfo) -> Model[T]:
        """Return the model object from the model info."""

    @abc.abstractmethod
    def to_create_schema(self) -> CreateModelSchema:
        """Return the create model schema object."""

    @abc.abstractmethod
    def resolve(self, base_path: t.Union[PathType, FS, None] = None) -> T:
        """Get the actual object of the model."""

    @t.overload
    def __get__(self, instance: None, owner: t.Type[t.Any]) -> t.Self: ...

    @t.overload
    def __get__(self, instance: t.Any, owner: t.Type[t.Any]) -> T: ...

    def __get__(self, instance: t.Any, owner: type) -> T | t.Self:
        if instance is None:
            return self
        if getattr(self, "_Model__resolved", None) is None:
            object.__setattr__(self, "_Model__resolved", self.resolve())
        return self.__resolved  # type: ignore[attr-defined]


@attrs.frozen
class BentoModel(Model[StoredModel]):
    """A model reference to a BentoML model.

    Args:
        tag (Tag): The model tag.

    Returns:
        Model: The bento model object.
    """

    tag: Tag = attrs.field(converter=Tag.from_taglike)

    @inject
    def _get_remote_model(
        self,
        rest_client: RestApiClient = Provide[BentoMLContainer.rest_api_client],
    ) -> ModelSchema | None:
        if self.tag.version in (None, "latest"):
            return rest_client.v1.get_latest_model(self.tag.name)
        else:
            return rest_client.v1.get_model(self.tag.name, self.tag.version)

    @property
    @inject
    def stored(
        self, model_store: ModelStore = Provide[BentoMLContainer.model_store]
    ) -> StoredModel | None:
        try:
            return model_store.get(self.tag)
        except NotFound:
            return None

    def to_info(self, alias: str | None = None) -> BentoModelInfo:
        stored = self.stored
        if stored is not None:
            return BentoModelInfo.from_bento_model(stored, alias)
        model = self._get_remote_model()
        if model is None:
            raise NotFound(f"Model {self.tag} not found either locally or remotely.")
        tag = Tag(self.tag.name, model.version)
        return BentoModelInfo(
            tag=tag,
            alias=alias,
            module=model.manifest.module,
            creation_time=model.created_at,
        )

    @classmethod
    def from_info(cls, info: BentoModelInfo) -> BentoModel:
        return cls(tag=info.tag)

    @inject
    def resolve(
        self,
        base_path: t.Union[PathType, FS, None] = None,
        global_model_store: ModelStore = Provide[BentoMLContainer.model_store],
    ) -> StoredModel:
        stored = self.stored
        if base_path is not None:
            model_store = ModelStore(base_path)
        else:
            model_store = global_model_store
        if stored is not None:
            if base_path is not None:
                copy_model(
                    stored.tag,
                    src_model_store=global_model_store,
                    target_model_store=model_store,
                )
            return stored
        cloud_client = BentoMLContainer.bentocloud_client.get()
        model = cloud_client.model.pull(self.tag, model_store=model_store)
        assert model is not None, "non-bentoml model"
        return model

    def to_create_schema(self) -> CreateModelSchema:
        """Return the create model schema object."""
        stored = self.stored
        if stored is None:
            raise ValueError("Not allowed to create a model without local store")
        info = stored.info
        labels = [
            LabelItemSchema(key=key, value=value) for key, value in info.labels.items()
        ]
        return CreateModelSchema(
            description="",
            version=stored.tag.version or "",
            build_at=info.creation_time,
            labels=labels,
            manifest=ModelManifestSchema(
                module=info.module,
                metadata=info.metadata,
                context=info.context.to_dict(),
                options=info.options.to_dict(),
                api_version=info.api_version,
                bentoml_version=info.context.bentoml_version,
                size_bytes=calc_dir_size(stored.path),
            ),
        )

    def __str__(self) -> str:
        return str(self.tag)



from __future__ import annotations

import hashlib
import logging
import platform
import subprocess
import sys
import typing as t
from pathlib import Path

import attrs

from bentoml._internal.bento.bento import ImageInfo
from bentoml._internal.bento.build_config import BentoBuildConfig
from bentoml._internal.configuration import DEFAULT_LOCK_PLATFORM
from bentoml._internal.configuration import get_bentoml_requirement
from bentoml._internal.configuration import get_debug_mode
from bentoml._internal.configuration import get_quiet_mode
from bentoml._internal.container.frontend.dockerfile import CONTAINER_METADATA
from bentoml._internal.container.frontend.dockerfile import CONTAINER_SUPPORTED_DISTROS
from bentoml._internal.utils.pkg import get_local_bentoml_dependency
from bentoml.exceptions import BentoMLConfigException
from bentoml.exceptions import BentoMLException

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

logger = logging.getLogger("bentoml.build")

DEFAULT_PYTHON_VERSION = f"{sys.version_info.major}.{sys.version_info.minor}"


@attrs.define
class Image:
    """A class defining the environment requirements for bento."""

    base_image: str
    python_version: str = DEFAULT_PYTHON_VERSION
    commands: t.List[str] = attrs.field(factory=list)
    lock_python_packages: bool = True
    python_requirements: str = ""
    post_commands: t.List[str] = attrs.field(factory=list)
    scripts: t.Dict[str, str] = attrs.field(factory=dict, init=False)
    _after_pip_install: bool = attrs.field(init=False, default=False, repr=False)

    def requirements_file(self, file_path: str) -> t.Self:
        """Add a requirements file to the image. Supports chaining call.

        Example:

        .. code-block:: python

            image = Image("debian:latest").requirements_file("requirements.txt")
        """
        self.python_requirements += Path(file_path).read_text()
        self._after_pip_install = True
        return self

    def pyproject_toml(self, file_path: str) -> t.Self:
        """Add a pyproject.toml to the image. Supports chaining call.

        Example:

        .. code-block:: python

            image = Image("debian:latest").pyproject_toml("pyproject.toml")
        """
        with Path(file_path).open("rb") as f:
            pyproject_toml = tomllib.load(f)
        dependencies = pyproject_toml.get("project", {}).get("dependencies", {})
        self.python_packages(*dependencies)
        return self

    def python_packages(self, *packages: str) -> t.Self:
        """Add python dependencies to the image. Supports chaining call.

        Example:

        .. code-block:: python

            image = Image("debian:latest")\
                .python_packages("numpy", "pandas")\
                .requirements_file("requirements.txt")
        """
        if not packages:
            raise BentoMLConfigException("No packages provided")
        self.python_requirements += "\n".join(packages) + "\n"
        self._after_pip_install = True
        return self

    def run(self, command: str) -> t.Self:
        """Add a command to the image. Supports chaining call.

        Example:

        .. code-block:: python

            image = Image("debian:latest").run("echo 'Hello, World!'")
        """
        commands = self.post_commands if self._after_pip_install else self.commands
        commands.append(command)
        return self

    def run_script(self, script: str) -> t.Self:
        """Run a script in the image. Supports chaining call.

        Example:

        .. code-block:: python

            image = Image("debian:latest").run_script("script.sh")
        """
        commands = self.post_commands if self._after_pip_install else self.commands
        script = Path(script).resolve().as_posix()
        # Files under /env/docker will be copied into the env image layer
        target_script = (
            f"./env/docker/script__{hashlib.md5(script.encode()).hexdigest()}"
        )
        commands.append(f"chmod +x {target_script} && {target_script}")
        self.scripts[script] = target_script
        return self

    def freeze(self, platform_: str | None = None) -> ImageInfo:
        """Freeze the image to an ImageInfo object for build."""
        python_requirements = self._freeze_python_requirements(platform_)
        return ImageInfo(
            base_image=self.base_image,
            python_version=self.python_version,
            commands=["export UV_COMPILE_BYTECODE=1", *self.commands],
            python_requirements=python_requirements,
            post_commands=self.post_commands,
            scripts=self.scripts,
        )

    def _freeze_python_requirements(self, platform_: str | None = None) -> str:
        from tempfile import TemporaryDirectory

        from pip_requirements_parser import RequirementsFile

        from bentoml._internal.configuration import get_uv_command

        with TemporaryDirectory(prefix="bento-reqs-") as parent:
            requirements_in = Path(parent).joinpath("requirements.in")
            requirements_in.write_text(self.python_requirements)
            # XXX: RequirementsFile.from_string() does not work due to bugs
            requirements_file = RequirementsFile.from_file(str(requirements_in))
            has_bentoml_req = any(
                req.name and req.name.lower() == "bentoml" and req.link is not None
                for req in requirements_file.requirements
            )
            with requirements_in.open("w") as f:
                f.write(requirements_file.dumps(preserve_one_empty_line=True))
                if not has_bentoml_req:
                    req = get_bentoml_requirement() or get_local_bentoml_dependency()
                    f.write(f"{req}\n")
            if not self.lock_python_packages:
                return requirements_in.read_text()
            lock_args = [
                str(requirements_in),
                "--allow-unsafe",
                "--no-header",
                f"--output-file={requirements_in.with_suffix('.lock')}",
                "--emit-index-url",
                "--emit-find-links",
                "--no-annotate",
            ]
            if get_debug_mode():
                lock_args.append("--verbose")
            else:
                lock_args.append("--quiet")
            logger.info("Locking PyPI package versions.")
            if platform_:
                lock_args.extend(["--python-platform", platform_])
            elif platform.system() != "Linux" or platform.machine() != "x86_64":
                logger.info(
                    "Locking packages for %s. Pass `--platform` option to specify the platform.",
                    DEFAULT_LOCK_PLATFORM,
                )
                lock_args.extend(["--python-platform", DEFAULT_LOCK_PLATFORM])
            cmd = [*get_uv_command(), "pip", "compile", *lock_args]
            try:
                subprocess.check_call(
                    cmd,
                    text=True,
                    stderr=subprocess.DEVNULL if get_quiet_mode() else None,
                    cwd=parent,
                )
            except subprocess.CalledProcessError as e:
                raise BentoMLException(f"Failed to lock PyPI packages: {e}") from None
            locked_requirements = (  # uv doesn't preserve global option lines, add them here
                "\n".join(option.dumps() for option in requirements_file.options)
            )
            if locked_requirements:
                locked_requirements += "\n"
            locked_requirements += requirements_in.with_suffix(".lock").read_text()
            return locked_requirements


@attrs.define
class PythonImage(Image):
    base_image: str = ""
    distro: str = "debian"
    _original_base_image: str = attrs.field(init=False, default="")

    def __attrs_post_init__(self) -> None:
        self._original_base_image = self.base_image
        if not self.base_image:
            if self.distro not in CONTAINER_METADATA:
                raise BentoMLConfigException(
                    f"Unsupported distro: {self.distro}, expected one of {CONTAINER_SUPPORTED_DISTROS}"
                )
            python_version = self.python_version
            if self.distro in ("ubi8",):
                python_version = python_version.replace(".", "")
            distro_config = CONTAINER_METADATA[self.distro]
            self.base_image = distro_config["python"]["image"].format(
                spec_version=python_version
            )
            self.commands.insert(0, distro_config["default_install_command"])

    def system_packages(self, *packages: str) -> t.Self:
        if self._original_base_image:
            raise BentoMLConfigException(
                "system_packages() can only be used in default base image"
            )
        if self.distro not in CONTAINER_METADATA:
            raise BentoMLConfigException(
                f"Unsupported distro: {self.distro}, expected one of {CONTAINER_SUPPORTED_DISTROS}"
            )
        self.commands.append(
            CONTAINER_METADATA[self.distro]["install_command"].format(
                packages=" ".join(packages)
            )
        )
        return self


def get_image_from_build_config(
    build_config: BentoBuildConfig, build_ctx: str
) -> Image | None:
    from bentoml._internal.utils.filesystem import resolve_user_filepath

    if not build_config.conda.is_empty():
        logger.warning(
            "conda options are not supported by bento v2, fallback to bento v1"
        )
        return None
    docker_options = build_config.docker
    if docker_options.cuda_version is not None:
        logger.warning(
            "docker.cuda_version is not supported by bento v2, fallback to bento v1"
        )
        return None
    if docker_options.dockerfile_template is not None:
        logger.warning(
            "docker.dockerfile_template is not supported by bento v2, fallback to bento v1"
        )
        return None
    image_params = {}
    if docker_options.base_image is not None:
        image_params["base_image"] = docker_options.base_image
    if docker_options.distro is not None:
        image_params["distro"] = docker_options.distro
    if docker_options.python_version is not None:
        image_params["python_version"] = docker_options.python_version
    image = PythonImage(**image_params)
    if docker_options.system_packages:
        image.system_packages(*docker_options.system_packages)

    python_options = build_config.python.with_defaults()
    if python_options.wheels:
        logger.warning(
            "python.wheels is not supported by bento v2, fallback to bento v1"
        )
        return None
    if python_options.pack_git_packages:
        logger.warning(
            "python.pack_git_packages is not supported by bento v2, fallback to bento v1"
        )
        return None
    image.lock_python_packages = python_options.lock_packages
    if python_options.index_url:
        image.python_packages(f"--index-url {python_options.index_url}")
    if python_options.no_index:
        image.python_packages("--no-index")
    if python_options.trusted_host:
        image.python_packages(
            *(f"--trusted-host {h}" for h in python_options.trusted_host)
        )
    if python_options.extra_index_url:
        image.python_packages(
            *(f"--extra-index-url {url}" for url in python_options.extra_index_url)
        )
    if python_options.find_links:
        image.python_packages(
            *(f"--find-links {link}" for link in python_options.find_links)
        )
    if python_options.requirements_txt:
        image.requirements_file(
            resolve_user_filepath(python_options.requirements_txt, build_ctx)
        )
    if python_options.packages:
        image.python_packages(*python_options.packages)
    if docker_options.setup_script:
        image.run_script(docker_options.setup_script)
    return image



from __future__ import annotations

import typing as t
from enum import Enum
from typing import TYPE_CHECKING

import attr

from ...bento.bento import BentoServiceInfo
from ...bento.bento import ImageInfo
from ...bento.build_config import BentoEnvSchema
from ...cloud.schemas.utils import dict_options_converter
from ...tag import Tag

time_format = "%Y-%m-%d %H:%M:%S.%f"

T = t.TypeVar("T")


class ResourceType(Enum):
    USER = "user"
    ORG = "organization"
    CLUSTER = "cluster"
    HostCluster = "host_cluster"
    BENTO_REPOSITORY = "bento_repository"
    BENTO = "bento"
    MODEL_REPOSITORY = "model_repository"
    MODEL = "model"
    DEPLOYMENT = "deployment"
    DEPLOYMENT_REVISION = "deployment_revision"
    TERMINAL_RECORD = "terminal_record"
    LABEL = "label"
    API_TOKEN = "api_token"
    YATAI_COMPONENT = "yatai_component"
    LimitGroup = "limit_group"
    ResourceInstance = "resource_instance"


class BentoImageBuildStatus(Enum):
    PENDING = "pending"
    BUILDING = "building"
    SUCCESS = "success"
    FAILED = "failed"


class UploadStatus(Enum):
    PENDING = "pending"
    BUILDING = "uploading"
    SUCCESS = "success"
    FAILED = "failed"


@attr.define
class BentoApiSchema:
    route: str
    doc: str
    input: str
    output: str


@attr.define
class BentoRunnerResourceSchema:
    cpu: t.Optional[t.Any]
    nvidia_gpu: t.Optional[t.Any]
    custom_resources: t.Optional[t.Any]


@attr.define
class BentoRunnerSchema:
    name: str
    runnable_type: t.Optional[str]
    models: t.Optional[t.List[str]]
    resource_config: t.Optional[BentoRunnerResourceSchema]


@attr.define
class BentoManifestSchema:
    service: str
    bentoml_version: str = attr.field(eq=False)
    size_bytes: int = attr.field(eq=False)
    entry_service: str = ""
    name: t.Optional[str] = None
    apis: t.Dict[str, BentoApiSchema] = attr.field(factory=dict)
    models: t.List[str] = attr.field(factory=list, eq=False)
    runners: t.Optional[t.List[BentoRunnerSchema]] = attr.field(factory=list)
    services: t.List[BentoServiceInfo] = attr.field(factory=dict)
    envs: t.List[BentoEnvSchema] = attr.field(factory=list)
    schema: t.Dict[str, t.Any] = attr.field(factory=dict)
    version: t.Optional[str] = attr.field(default=None, eq=False)
    dev: bool = attr.field(default=False, eq=False)
    image: t.Optional[ImageInfo] = attr.field(default=None, eq=False)
    spec: int = attr.field(default=1)

    @property
    def tag(self) -> Tag:
        return Tag(self.name, self.version)


if TYPE_CHECKING:
    TransmissionStrategy = t.Literal["presigned_url", "proxy"]
else:
    TransmissionStrategy = str


@attr.define
class ModelManifestSchema:
    module: str
    api_version: str
    bentoml_version: str
    size_bytes: int
    metadata: t.Dict[str, t.Any] = attr.field(factory=dict)
    context: t.Dict[str, t.Any] = attr.field(factory=dict)
    options: t.Dict[str, t.Any] = attr.field(factory=dict)


@attr.define
class DeploymentTargetCanaryRule:
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    type: str
    weight: int
    header: str
    cookie: str
    header_value: str


@attr.define
class ApiServerBentoDeploymentOverrides:
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    monitorExporter: t.Optional[t.Dict[str, t.Any]] = attr.field(default=None)
    extraPodMetadata: t.Optional[t.Dict[str, t.Any]] = attr.field(default=None)
    extraPodSpec: t.Optional[t.Dict[str, t.Any]] = attr.field(default=None)


@attr.define
class ApiServerBentoFunctionOverrides:
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    annotations: t.Optional[t.Dict[str, str]] = attr.field(default=None)
    monitorExporter: t.Optional[t.Dict[str, t.Any]] = attr.field(default=None)
    extraPodMetadata: t.Optional[t.Dict[str, t.Any]] = attr.field(default=None)
    extraPodSpec: t.Optional[t.Dict[str, t.Any]] = attr.field(default=None)


@attr.define
class RunnerBentoFunctionOverrides:
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    annotations: t.Optional[t.Dict[str, str]] = attr.field(default=None)
    extraPodMetadata: t.Optional[t.Dict[str, t.Any]] = attr.field(default=None)
    extraPodSpec: t.Optional[t.Dict[str, t.Any]] = attr.field(default=None)


@attr.define
class RunnerBentoDeploymentOverrides:
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    extraPodMetadata: t.Optional[t.Dict[str, t.Any]] = attr.field(default=None)
    extraPodSpec: t.Optional[t.Dict[str, t.Any]] = attr.field(default=None)


@attr.define
class BentoRequestOverrides:
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    imageBuildTimeout: int = attr.field(default=None)
    imageBuilderExtraPodMetadata: t.Optional[t.Dict[str, t.Any]] = attr.field(
        default=None
    )
    imageBuilderExtraPodSpec: t.Optional[t.Dict[str, t.Any]] = attr.field(default=None)
    imageBuilderExtraContainerEnv: t.Optional[t.List[t.Dict[str, t.Any]]] = attr.field(
        default=None
    )
    imageBuilderContainerResources: t.Optional[t.Dict[str, t.Any]] = attr.field(
        default=None
    )
    dockerConfigJsonSecretName: t.Optional[str] = attr.field(default=None)
    downloaderContainerEnvFrom: t.Optional[t.Dict[str, t.Any]] = attr.field(
        default=None
    )


@attr.define
class LabelItemSchema:
    key: str
    value: str


@attr.define
class EnvItemSchema:
    name: str
    value: str


@attr.define
class HPAMetric:
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    type: str
    value: t.Any  # resource.Quantity


@attr.define
class HPAPolicy:
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    metrics: t.Optional[t.List[HPAMetric]] = attr.field(default=None)
    scale_down_behavior: t.Optional[str] = attr.field(default=None)
    scale_up_behavior: t.Optional[str] = attr.field(default=None)


@attr.define
class DeploymentTargetHPAConf:
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    min_replicas: t.Optional[int] = attr.field(default=None)
    max_replicas: t.Optional[int] = attr.field(default=None)
    policy: t.Optional[HPAPolicy] = attr.field(default=None)


@attr.define
class DeploymentTargetResourceItem:
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    cpu: t.Optional[str] = attr.field(default=None)
    memory: t.Optional[str] = attr.field(default=None)
    gpu: t.Optional[str] = attr.field(default=None)
    custom: t.Optional[t.Dict[str, str]] = attr.field(default=None)


@attr.define
class DeploymentTargetResources:
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    requests: t.Optional[DeploymentTargetResourceItem] = attr.field(default=None)
    limits: t.Optional[DeploymentTargetResourceItem] = attr.field(default=None)


@attr.define
class RequestQueueConfig:
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    enabled: t.Optional[bool] = attr.field(default=None)
    max_consume_concurrency: t.Optional[int] = attr.field(default=None)


@attr.define
class TrafficControlConfig:
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    timeout: t.Optional[str] = attr.field(default=None)
    request_queue: t.Optional[RequestQueueConfig] = attr.field(default=None)


class DeploymentStrategy(Enum):
    RollingUpdate = "RollingUpdate"
    Recreate = "Recreate"
    RampedSlowRollout = "RampedSlowRollout"
    BestEffortControlledRollout = "BestEffortControlledRollout"


@attr.define
class DeploymentTargetRunnerConfig:
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    resource_instance: t.Optional[str] = attr.field(default=None)
    resources: t.Optional[DeploymentTargetResources] = attr.field(default=None)
    hpa_conf: t.Optional[DeploymentTargetHPAConf] = attr.field(default=None)
    envs: t.Optional[t.List[t.Optional[LabelItemSchema]]] = attr.field(default=None)
    enable_stealing_traffic_debug_mode: t.Optional[bool] = attr.field(default=None)
    enable_debug_mode: t.Optional[bool] = attr.field(default=None)
    enable_debug_pod_receive_production_traffic: t.Optional[bool] = attr.field(
        default=None
    )
    deployment_strategy: t.Optional[str] = attr.field(default=None)
    bento_deployment_overrides: t.Optional[RunnerBentoDeploymentOverrides] = attr.field(
        default=None
    )
    bento_function_overrides: t.Optional[RunnerBentoFunctionOverrides] = attr.field(
        default=None
    )
    traffic_control: t.Optional[TrafficControlConfig] = attr.field(default=None)
    deployment_cold_start_wait_timeout: t.Optional[int] = attr.field(default=None)


@attr.define
class DeploymentTargetConfig:
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    resources: t.Optional[DeploymentTargetResources] = attr.field(
        default=None, converter=dict_options_converter(DeploymentTargetResources)
    )
    kubeResourceUid: str = attr.field(default="")  # empty str
    kubeResourceVersion: str = attr.field(default="")
    resource_instance: t.Optional[str] = attr.field(default=None)
    hpa_conf: t.Optional[DeploymentTargetHPAConf] = attr.field(default=None)
    envs: t.Optional[t.List[t.Optional[LabelItemSchema]]] = attr.field(default=None)
    runners: t.Optional[t.Dict[str, DeploymentTargetRunnerConfig]] = attr.field(
        default=None
    )
    access_control: t.Optional[str] = attr.field(default=None)
    enable_ingress: t.Optional[bool] = attr.field(default=None)  # false for enables
    enable_stealing_traffic_debug_mode: t.Optional[bool] = attr.field(default=None)
    enable_debug_mode: t.Optional[bool] = attr.field(default=None)
    enable_debug_pod_receive_production_traffic: t.Optional[bool] = attr.field(
        default=None
    )
    deployment_strategy: t.Optional[str] = attr.field(default=None)  # Specific
    bento_deployment_overrides: t.Optional[ApiServerBentoDeploymentOverrides] = (
        attr.field(default=None)
    )
    bento_request_overrides: t.Optional[BentoRequestOverrides] = attr.field(
        default=None
    )  # Put into image builder
    bento_function_overrides: t.Optional[ApiServerBentoFunctionOverrides] = attr.field(
        default=None
    )
    traffic_control: t.Optional[TrafficControlConfig] = attr.field(default=None)
    deployment_cold_start_wait_timeout: t.Optional[int] = attr.field(default=None)


@attr.define
class ExtraDeploymentOverrides:
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    bento_function_overrides: t.Optional[ApiServerBentoFunctionOverrides] = attr.field(
        default=None
    )
    bento_request_overrides: t.Optional[BentoRequestOverrides] = attr.field(
        default=None
    )


@attr.define
class DeploymentServiceConfig:
    __omit_if_default__ = True
    __forbid_extra_keys__ = False
    instance_type: t.Optional[str] = attr.field(default=None)
    scaling: t.Optional[DeploymentTargetHPAConf] = attr.field(default=None)
    envs: t.Optional[t.List[t.Optional[EnvItemSchema]]] = attr.field(default=None)
    deployment_strategy: t.Optional[str] = attr.field(default=None)
    extras: t.Optional[ExtraDeploymentOverrides] = attr.field(default=None)
    cold_start_timeout: t.Optional[int] = attr.field(default=None)
    config_overrides: t.Optional[t.Dict[str, t.Any]] = attr.field(factory=dict)


class DeploymentStatus(Enum):
    Unknown = "unknown"
    NonDeployed = "non-deployed"
    Running = "running"
    Unhealthy = "unhealthy"
    Failed = "failed"
    Deploying = "deploying"
    Terminating = "terminating"
    Terminated = "terminated"
    ImageBuilding = "image-building"
    ImageBuildFailed = "image-build-failed"
    ImageBuildSucceeded = "image-build-succeeded"
    ScaledToZero = "scaled-to-zero"


@attr.define
class ResourceInstanceConfigSchema:
    group: str
    resources: DeploymentTargetResources
    price: str
    node_selectors: t.Optional[t.Dict[str, str]] = attr.field(factory=dict)
    gpu_config: t.Optional[ResourceInstanceGPUConfigSchema] = attr.field(default=None)


@attr.define
class ResourceInstanceGPUConfigSchema:
    type: str
    memory: str


class DeploymentRevisionStatus(Enum):
    ACTIVE = "active"
    INACTIVE = "inactive"



from __future__ import annotations

import typing as t
from pathlib import Path

from bentoml.exceptions import MissingDependencyException

from .decorators import asgi_app

R = t.TypeVar("R")

try:
    import gradio as gr
    from gradio import Blocks
except ImportError:  # pragma: no cover
    raise MissingDependencyException(
        """'gradio' is required by 'bentoml.gradio', run 'pip install -U gradio' to install gradio.""",
    )


def mount_gradio_app(blocks: Blocks, path: str, name: str = "gradio_ui"):
    """Mount a Gradio app to a BentoML service.

    Args:
        blocks: The Gradio blocks to be mounted.
        path: The URL path to mount the Gradio app.
        name: The name of the Gradio app.

    Example:

    Both of the following examples are allowed::

        @bentoml.gradio.mount_gradio_app(blocks)
        @bentoml.service()
        class MyService: ...

        @bentoml.service()
        @bentoml.gradio.mount_gradio_app(blocks)
        class MyService: ...

    """
    from _bentoml_sdk.service import Service
    from bentoml._internal import server

    favicon_path = str(
        Path(server.__file__).parent / "static_content" / "favicon-light-32x32.png"
    )
    assert path.startswith("/"), "Routed paths must start with '/'"
    path = path.rstrip("/")

    def decorator(obj: R) -> R:
        blocks.dev_mode = False
        blocks.show_error = True
        blocks.max_file_size = None
        blocks.validate_queue_settings()
        blocks.root_path = path
        blocks.favicon_path = favicon_path
        gradio_app = gr.routes.App.create_app(blocks, app_kwargs={"root_path": path})
        asgi_app(gradio_app, path=path, name=name)(obj)

        # @bentoml.service() decorator returns a wrapper instead of the original class
        # Check if the object is an instance of Service
        if isinstance(obj, Service):
            # For scenario:
            #
            # @bentoml.gradio.mount_gradio_app(..)
            # @bentoml.service(..)
            # class MyService: ...
            #
            # If the Service instance is already created, mount the ASGI app immediately
            target = obj.inner
        else:
            # For scenario:
            #
            # @bentoml.service(..)
            # @bentoml.gradio.mount_gradio_app(..)
            # class MyService: ...
            #
            # If the Service instance is not yet created, mark the Gradio app info
            # for later mounting during Service instance initialization
            target = obj

        # Store Gradio app information for ASGI app mounting and startup event callback
        gradio_apps = getattr(target, "__bentoml_gradio_apps__", [])
        gradio_apps.append((gradio_app, path, name))
        setattr(target, "__bentoml_gradio_apps__", gradio_apps)

        return obj

    return decorator



from __future__ import annotations

import typing as t
from contextlib import contextmanager

import attrs
from rich import get_console
from rich.console import Group
from rich.live import Live
from rich.panel import Panel
from rich.progress import BarColumn
from rich.progress import DownloadColumn
from rich.progress import Progress
from rich.progress import SpinnerColumn
from rich.progress import TaskID
from rich.progress import TextColumn
from rich.progress import TimeElapsedColumn
from rich.progress import TimeRemainingColumn
from rich.progress import TransferSpeedColumn

if t.TYPE_CHECKING:
    from rich.console import Console
    from rich.console import ConsoleOptions
    from rich.console import RenderResult

FILE_CHUNK_SIZE = 100 * 1024 * 1024  # 100Mb
UPLOAD_RETRY_COUNT = 3


@attrs.define
class CallbackIOWrapper(t.IO[bytes]):
    file: t.IO[bytes]
    read_cb: t.Callable[[int], None] | None = None
    write_cb: t.Callable[[int], None] | None = None
    start: int | None = None
    end: int | None = None

    def __attrs_post_init__(self) -> None:
        self.reset()

    def reset(self) -> int:
        read = self.tell() - (self.start or 0)
        self.file.seek(self.start or 0, 0)
        return read

    def seek(self, offset: int, whence: int = 0) -> int:
        if whence == 2 and self.end is not None:
            length = self.file.seek(self.end, 0)
        else:
            length = self.file.seek(offset, whence)
        return length - (self.start or 0)

    def tell(self) -> int:
        return self.file.tell()

    def fileno(self) -> int:
        # Raise OSError to prevent access to the underlying file descriptor
        raise OSError("fileno")

    def __getattr__(self, name: str) -> t.Any:
        return getattr(self.file, name)

    def read(self, size: int = -1) -> bytes:
        pos = self.tell()
        if self.end is not None:
            if size < 0 or size > self.end - pos:
                size = self.end - pos
        res = self.file.read(size)
        if self.read_cb is not None:
            self.read_cb(len(res))
        return res

    def write(self, data: bytes) -> t.Any:  # type: ignore  # python buffer types are too new and seem to not support something like Buffer+Sized as of now
        res = super().write(data)
        if self.write_cb is not None:
            if hasattr(data, "__len__"):
                self.write_cb(len(data))
        return res

    def __iter__(self) -> t.Iterator[bytes]:
        return iter(self.file)


class Spinner:
    """A UI component that renders as follows:
    ┌────────────────────────────────────┐
    │ This is log                        │log
    │ This is another log                │
    │                                    │
    │ 00:45 Loading ...                  │spin
    └────────────────────────────────────┘
    Processing -----                00:05 progress

    Use it as a context manager to start the live updating.
    """

    def __init__(self, console: Console | None = None) -> None:
        self.console = console or get_console()
        self.transmission_progress = Progress(
            TextColumn("[bold blue]{task.description}", justify="right"),
            BarColumn(bar_width=None),
            "[progress.percentage]{task.percentage:>3.1f}%",
            "•",
            DownloadColumn(),
            "•",
            TransferSpeedColumn(),
            "•",
            TimeRemainingColumn(),
            console=self.console,
        )

        self._logs: list[str] = []
        self._spinner_progress = Progress(
            TextColumn("  "),
            TimeElapsedColumn(),
            TextColumn("[bold purple]{task.description}"),
            SpinnerColumn("simpleDots"),
            console=self.console,
        )
        self._spinner_task_id: t.Optional[TaskID] = None
        self._live = Live(self, console=self.console)
        self._start_count = 0

    @contextmanager
    def spin(self, text: str) -> t.Generator[TaskID, None, None]:
        """Create a spinner as a context manager."""
        task_id = self.update(text, new=True)
        try:
            yield task_id
        finally:
            self._spinner_progress.remove_task(task_id)
            if self._spinner_task_id == task_id:
                self._spinner_task_id = None

    def update(self, text: str, new: bool = False) -> TaskID:
        """Update the spin text."""
        if self._spinner_task_id is None or new:
            task_id = self._spinner_progress.add_task(text)
            if self._spinner_task_id is None:
                self._spinner_task_id = task_id
        else:
            task_id = self._spinner_task_id
            self._spinner_progress.update(task_id, description=text)
        return task_id

    def __rich_console__(
        self, console: "Console", options: "ConsoleOptions"
    ) -> RenderResult:
        yield Panel(Group(*self._logs, self._spinner_progress))
        yield self.transmission_progress

    def start(self) -> None:
        """Start live updating."""
        self._start_count += 1
        self._live.start()

    def stop(self) -> None:
        """Stop live updating."""
        self._start_count -= 1
        if self._start_count > 0:
            return
        if self._spinner_task_id is not None:
            self._spinner_progress.remove_task(self._spinner_task_id)
            self._spinner_task_id = None
        self._live.stop()

    def log(self, message: str) -> None:
        """Add a log message line."""
        self._logs.append(message)

    def __enter__(self) -> Spinner:
        self.start()
        return self

    def __exit__(self, *_: t.Any) -> None:
        self.stop()



from __future__ import annotations

import io
import logging
import os
import typing as t
from functools import lru_cache

from multipart.multipart import parse_options_header
from starlette.datastructures import UploadFile
from starlette.requests import Request
from starlette.responses import Response

from ...exceptions import BadInput
from ...exceptions import BentoMLException
from ...exceptions import InvalidArgument
from ...grpc.utils import import_generated_stubs
from ..service.openapi import SUCCESS_DESCRIPTION
from ..service.openapi.specification import MediaType
from ..service.openapi.specification import Schema
from ..types import FileLike
from ..utils.filesystem import resolve_user_filepath
from ..utils.http import set_cookies
from .base import IODescriptor

logger = logging.getLogger(__name__)

if t.TYPE_CHECKING:
    from bentoml.grpc.v1 import service_pb2 as pb
    from bentoml.grpc.v1alpha1 import service_pb2 as pb_v1alpha1

    from ..context import ServiceContext as Context
    from .base import OpenAPIResponse

    FileKind: t.TypeAlias = t.Literal["binaryio", "textio"]
else:
    pb, _ = import_generated_stubs("v1")
    pb_v1alpha1, _ = import_generated_stubs("v1alpha1")

FileType = t.Union[io.IOBase, t.IO[bytes], FileLike[bytes]]


class File(
    IODescriptor[FileType], descriptor_id="bentoml.io.File", proto_fields=("file",)
):
    """
    :obj:`File` defines API specification for the inputs/outputs of a Service, where either
    inputs will be converted to or outputs will be converted from file-like objects as
    specified in your API function signature.

    A sample ViT service:

    .. code-block:: python
       :caption: `service.py`

       from __future__ import annotations

       import io
       from typing import TYPE_CHECKING
       from typing import Any

       import bentoml
       from bentoml.io import File

       if TYPE_CHECKING:
           from numpy.typing import NDArray

       runner = bentoml.tensorflow.get('image-classification:latest').to_runner()

       svc = bentoml.Service("vit-pdf-classifier", runners=[runner])

       @svc.api(input=File(), output=NumpyNdarray(dtype="float32"))
       async def predict(input_pdf: io.BytesIO[Any]) -> NDArray[Any]:
           return await runner.async_run(input_pdf)

    Users then can then serve this service with :code:`bentoml serve`:

    .. code-block:: bash

        % bentoml serve ./service.py:svc --reload

    Users can then send requests to the newly started services with any client:

    .. tab-set::

        .. tab-item:: Bash

            .. code-block:: bash

               % curl -H "Content-Type: multipart/form-data" \\
                      -F 'fileobj=@test.pdf;type=application/pdf' \\
                      http://0.0.0.0:3000/predict

        .. tab-item:: Python

           .. code-block:: python
              :caption: `request.py`

              import requests

              requests.post(
                  "http://0.0.0.0:3000/predict",
                  files = {"upload_file": open('test.pdf', 'rb')},
                  headers = {"content-type": "multipart/form-data"}
              ).text

    Args:
        kind: The kind of file-like object to be used. Currently, the only accepted value is ``binaryio``.
        mime_type: Return MIME type of the :code:`starlette.response.Response`, only available when used as output descriptor

    Returns:
        :obj:`File`: IO Descriptor that represents file-like objects.

    """

    def __new__(
        cls, kind: FileKind = "binaryio", mime_type: str | None = None, **kwargs: t.Any
    ) -> File:
        if kind == "binaryio":
            res = super().__new__(BytesIOFile, **kwargs)
        else:
            raise ValueError(f"invalid File kind '{kind}'")
        res._mime_type = mime_type
        return res

    def _from_sample(self, sample: FileType | str) -> FileType:
        """
        Create a :class:`~bentoml._internal.io_descriptors.file.File` IO Descriptor from given inputs.

        Args:
            sample: Given File-like object, or a path to a file.
            kind: The kind of file-like object to be used. Currently, the only accepted value is ``binaryio``.
            mime_type: Optional MIME type for the descriptor. If not provided, ``from_sample``
                       will try to infer the MIME type from the file extension.

        Returns:
            :class:`~bentoml._internal.io_descriptors.file.File`: IODescriptor from given users inputs.

        Example:

        .. code-block:: python
           :caption: `service.py`

           from __future__ import annotations

           import bentoml
           from typing import Any
           from bentoml.io import File

           input_spec = File.from_sample("/path/to/file.pdf")
           @svc.api(input=input_spec, output=File())
           async def predict(input: t.IO[t.Any]) -> t.IO[t.Any]:
               return await runner.async_run(input)

        Raises:
            :class:`MissingDependencyException`: If ``filetype`` is not installed.
            ValueError: If the MIME type cannot be inferred automatically.
            :class:`BadInput`: Any other errors that may occur during the process of figure out the MIME type.
        """
        import mimetypes

        if isinstance(sample, t.IO):
            sample = FileLike[bytes](sample, "<sample>")
        elif isinstance(sample, (str, os.PathLike)):
            p = resolve_user_filepath(sample, ctx=None)
            mime = mimetypes.guess_type(p)[0]
            self._mime_type = mime
            with open(p, "rb") as f:
                sample = FileLike[bytes](f, "<sample>")
        return sample

    @classmethod
    def from_spec(cls, spec: dict[str, t.Any]) -> t.Self:
        if "args" not in spec:
            raise InvalidArgument(f"Missing args key in File spec: {spec}")
        return cls(**spec["args"])

    def input_type(self) -> t.Type[t.Any]:
        return FileLike[bytes]

    def openapi_schema(self) -> Schema:
        return Schema(type="string", format="binary")

    def openapi_components(self) -> dict[str, t.Any] | None:
        pass

    def openapi_example(self):
        pass

    def openapi_request_body(self) -> dict[str, t.Any]:
        return {
            "content": {
                "*/*" if self._mime_type is None else self._mime_type: MediaType(
                    schema=self.openapi_schema()
                )
            },
            "required": True,
            "x-bentoml-io-descriptor": self.to_spec(),
        }

    def openapi_responses(self) -> OpenAPIResponse:
        return {
            "description": SUCCESS_DESCRIPTION,
            "content": {
                "*/*" if self._mime_type is None else self._mime_type: MediaType(
                    schema=self.openapi_schema()
                )
            },
            "x-bentoml-io-descriptor": self.to_spec(),
        }

    async def to_http_response(self, obj: FileType, ctx: Context | None = None):
        if isinstance(obj, bytes):
            body = obj
        else:
            body = obj.read()

        if ctx is not None:
            res = Response(
                body,
                headers=ctx.response.metadata,  # type: ignore (bad starlette types)
                status_code=ctx.response.status_code,
            )
            set_cookies(res, ctx.response.cookies)
        else:
            res = Response(
                body,
                headers={
                    "content-type": (
                        self._mime_type
                        if self._mime_type
                        else "application/octet-stream"
                    )
                },
            )
        return res

    async def to_proto(self, obj: FileType) -> pb.File:
        if isinstance(obj, bytes):
            body = obj
        else:
            body = obj.read()

        return pb.File(
            kind=(
                "appliction/octet-stream"
                if self._mime_type is None
                else self._mime_type
            ),
            content=body,
        )

    async def to_proto_v1alpha1(self, obj: FileType) -> pb_v1alpha1.File:
        if isinstance(obj, bytes):
            body = obj
        else:
            body = obj.read()

        try:
            kind = mimetype_to_filetype_pb_map()[
                (
                    "application/octet-stream"
                    if self._mime_type is None
                    else self._mime_type
                )
            ]
        except KeyError:
            raise BadInput(
                f"{self._mime_type} doesn't have a corresponding File 'kind'"
            ) from None

        return pb_v1alpha1.File(kind=kind, content=body)

    async def from_proto(self, field: pb.File | bytes) -> FileLike[bytes]:
        raise NotImplementedError

    async def from_http_request(self, request: Request) -> FileLike[bytes]:
        raise NotImplementedError


class BytesIOFile(File, descriptor_id=None):
    def to_spec(self) -> dict[str, t.Any]:
        return {
            "id": super().descriptor_id,
            "args": {
                "kind": "binaryio",
                "mime_type": self._mime_type,
            },
        }

    async def from_http_request(self, request: Request) -> FileLike[bytes]:
        content_type, _ = parse_options_header(request.headers["content-type"])
        if content_type.decode("utf-8") == "multipart/form-data":
            form = await request.form()
            found_mimes: t.List[str] = []
            val: t.Union[str, UploadFile]
            for val in form.values():  # type: ignore
                if isinstance(val, UploadFile):
                    found_mimes.append(val.content_type)  # type: ignore (bad starlette types)
                    if self._mime_type is None or val.content_type == self._mime_type:  # type: ignore (bad starlette types)
                        res = FileLike[bytes](val.file, val.filename)  # type: ignore (bad starlette types)
                        break
            else:
                if len(found_mimes) == 0:
                    raise BentoMLException("no File found in multipart form")
                else:
                    raise BentoMLException(
                        f"The File IO descriptor requires input to be of Content-Type '{self._mime_type}', got files with content types {', '.join(found_mimes)}"
                    )
            return res  # type: ignore
        if self.mime_type is None or content_type.decode("utf-8") == self._mime_type:
            body = await request.body()
            return FileLike[bytes](io.BytesIO(body), "<request body>")
        raise BentoMLException(
            f"File should have Content-Type '{self._mime_type}' or 'multipart/form-data', got {content_type} instead"
        )

    async def from_proto(
        self, field: pb.File | pb_v1alpha1.File | bytes
    ) -> FileLike[bytes]:
        # check if the request message has the correct field
        if isinstance(field, bytes):
            content = field
        elif isinstance(field, pb_v1alpha1.File):
            mapping = filetype_pb_to_mimetype_map()
            if field.kind:
                try:
                    mime_type = mapping[field.kind]
                    if self._mime_type is not None and mime_type != self._mime_type:
                        raise BadInput(
                            f"Inferred mime_type from 'kind' is '{mime_type}', while '{self!r}' is expecting '{self._mime_type}'",
                        )
                except KeyError:
                    raise BadInput(
                        f"{field.kind} is not a valid File kind. Accepted file kind: {[names for names, _ in pb_v1alpha1.File.FileType.items()]}",
                    ) from None
            content = field.content
            if not content:
                raise BadInput("Content is empty!") from None
            return FileLike[bytes](io.BytesIO(content), "<content>")
        else:
            assert isinstance(field, pb.File)
            if (
                field.kind is not None
                and self._mime_type is not None
                and field.kind != self._mime_type
            ):
                raise BadInput(
                    f"MIME type from 'kind' is '{field.kind}', while '{self!r}' is expecting '{self._mime_type}'",
                )
            content = field.content
            if not content:
                raise BadInput("Content is empty!") from None

        return FileLike[bytes](io.BytesIO(content), "<content>")


# v1alpha1 backward compatibility
@lru_cache(maxsize=1)
def filetype_pb_to_mimetype_map() -> dict[pb_v1alpha1.File.FileType.ValueType, str]:
    return {
        pb_v1alpha1.File.FILE_TYPE_CSV: "text/csv",
        pb_v1alpha1.File.FILE_TYPE_PLAINTEXT: "text/plain",
        pb_v1alpha1.File.FILE_TYPE_JSON: "application/json",
        pb_v1alpha1.File.FILE_TYPE_BYTES: "application/octet-stream",
        pb_v1alpha1.File.FILE_TYPE_PDF: "application/pdf",
        pb_v1alpha1.File.FILE_TYPE_PNG: "image/png",
        pb_v1alpha1.File.FILE_TYPE_JPEG: "image/jpeg",
        pb_v1alpha1.File.FILE_TYPE_GIF: "image/gif",
        pb_v1alpha1.File.FILE_TYPE_TIFF: "image/tiff",
        pb_v1alpha1.File.FILE_TYPE_BMP: "image/bmp",
        pb_v1alpha1.File.FILE_TYPE_WEBP: "image/webp",
        pb_v1alpha1.File.FILE_TYPE_SVG: "image/svg+xml",
    }


@lru_cache(maxsize=1)
def mimetype_to_filetype_pb_map() -> dict[str, pb_v1alpha1.File.FileType.ValueType]:
    return {v: k for k, v in filetype_pb_to_mimetype_map().items()}



from __future__ import annotations

from bentoml.exceptions import MissingDependencyException

try:
    import diffusers  # noqa
    import torch  # noqa
    from diffusers.loaders import LoraLoaderMixin as LoraLoaderMixin
    from diffusers.loaders import (
        TextualInversionLoaderMixin as TextualInversionLoaderMixin,
    )
except ImportError:  # pragma: no cover
    raise MissingDependencyException(
        "'diffusers' and 'transformers' are required in order to use module 'bentoml.diffusers_runners', install diffusers and its dependencies with 'pip install --upgrade diffusers transformers accelerate'. For more information, refer to https://github.com/huggingface/diffusers",
    )


from . import stable_diffusion as stable_diffusion
from . import stable_diffusion_xl as stable_diffusion_xl



from __future__ import annotations

import contextlib
import json
import logging
import os

from simple_di import Provide
from simple_di import inject

from ._internal.configuration.containers import BentoMLContainer
from ._internal.runner.runner import Runner
from .grpc.utils import LATEST_PROTOCOL_VERSION

logger = logging.getLogger(__name__)

SCRIPT_RUNNER = "bentoml_cli.worker.runner"
SCRIPT_API_SERVER = "bentoml_cli.worker.http_api_server"
SCRIPT_GRPC_API_SERVER = "bentoml_cli.worker.grpc_api_server"
SCRIPT_GRPC_PROMETHEUS_SERVER = "bentoml_cli.worker.grpc_prometheus_server"

API_SERVER = "api_server"
RUNNER = "runner"


@inject(squeeze_none=True)
def start_runner_server(
    bento_identifier: str,
    working_dir: str,
    runner_name: str,
    port: int | None = None,
    host: str | None = None,
    timeout: int | None = None,
    backlog: int = Provide[BentoMLContainer.api_server_config.backlog],
) -> None:
    """
    Experimental API for serving a BentoML runner.
    """
    from .serving import ensure_prometheus_dir

    env = {"PROMETHEUS_MULTIPROC_DIR": ensure_prometheus_dir()}

    from . import load
    from ._internal.utils import reserve_free_port
    from ._internal.utils.analytics import track_serve
    from ._internal.utils.circus import create_standalone_arbiter
    from .serving import create_watcher
    from .serving import find_triton_binary

    working_dir = os.path.realpath(os.path.expanduser(working_dir))
    svc = load(bento_identifier, working_dir=working_dir)
    timeout_args = ["--timeout", str(timeout)] if timeout else []

    from circus.sockets import CircusSocket  # type: ignore
    from circus.watcher import Watcher  # type: ignore

    watchers: list[Watcher] = []
    circus_socket_map: dict[str, CircusSocket] = {}

    # NOTE: We need to find and set model-repository args
    # to all TritonRunner instances (required from tritonserver if spawning multiple instances.)

    with contextlib.ExitStack() as port_stack:
        for runner in svc.runners:
            if runner.name == runner_name:
                if isinstance(runner, Runner):
                    if port is None:
                        port = port_stack.enter_context(reserve_free_port())
                    if host is None:
                        host = "127.0.0.1"
                    circus_socket_map[runner.name] = CircusSocket(
                        name=runner.name,
                        host=host,
                        port=port,
                        backlog=backlog,
                    )

                    watchers.append(
                        create_watcher(
                            name=f"{RUNNER}_{runner.name}",
                            args=[
                                "-m",
                                SCRIPT_RUNNER,
                                bento_identifier,
                                "--runner-name",
                                runner.name,
                                "--fd",
                                f"$(circus.sockets.{runner.name})",
                                "--working-dir",
                                working_dir,
                                "--no-access-log",
                                "--worker-id",
                                "$(circus.wid)",
                                "--worker-env-map",
                                json.dumps(runner.scheduled_worker_env_map),
                                *timeout_args,
                            ],
                            working_dir=working_dir,
                            numprocesses=runner.scheduled_worker_count,
                            env=env,
                        )
                    )
                    break
                else:
                    cli_args = runner.cli_args + [
                        (
                            f"--http-port={runner.protocol_address.split(':')[-1]}"
                            if runner.tritonserver_type == "http"
                            else f"--grpc-port={runner.protocol_address.split(':')[-1]}"
                        )
                    ]
                    watchers.append(
                        create_watcher(
                            name=f"tritonserver_{runner.name}",
                            cmd=find_triton_binary(),
                            args=cli_args,
                            use_sockets=False,
                            working_dir=working_dir,
                            numprocesses=1,
                            env=env,
                        )
                    )
                    break
        else:
            raise ValueError(
                f"Runner {runner_name} not found in the service: `{bento_identifier}`, "
                f"available runners: {[r.name for r in svc.runners]}"
            )

    arbiter = create_standalone_arbiter(
        watchers=watchers,
        sockets=list(circus_socket_map.values()),
    )
    with track_serve(svc, production=True, component=RUNNER):
        arbiter.start(
            cb=lambda _: logger.info(  # type: ignore
                'Starting RunnerServer from "%s" running on http://%s:%s (Press CTRL+C to quit)',
                bento_identifier,
                host,
                port,
            ),
        )


@inject(squeeze_none=True)
def start_http_server(
    bento_identifier: str,
    runner_map: dict[str, str],
    working_dir: str,
    port: int = Provide[BentoMLContainer.api_server_config.port],
    host: str = Provide[BentoMLContainer.api_server_config.host],
    backlog: int = Provide[BentoMLContainer.api_server_config.backlog],
    api_workers: int = Provide[BentoMLContainer.api_server_workers],
    timeout: int | None = None,
    ssl_certfile: str | None = Provide[BentoMLContainer.ssl.certfile],
    ssl_keyfile: str | None = Provide[BentoMLContainer.ssl.keyfile],
    ssl_keyfile_password: str | None = Provide[BentoMLContainer.ssl.keyfile_password],
    ssl_version: int | None = Provide[BentoMLContainer.ssl.version],
    ssl_cert_reqs: int | None = Provide[BentoMLContainer.ssl.cert_reqs],
    ssl_ca_certs: str | None = Provide[BentoMLContainer.ssl.ca_certs],
    ssl_ciphers: str | None = Provide[BentoMLContainer.ssl.ciphers],
    timeout_keep_alive: int | None = None,
    timeout_graceful_shutdown: int | None = None,
) -> None:
    from .serving import ensure_prometheus_dir

    env = {"PROMETHEUS_MULTIPROC_DIR": ensure_prometheus_dir()}

    from circus.sockets import CircusSocket
    from circus.watcher import Watcher

    from . import load
    from ._internal.utils.analytics import track_serve
    from ._internal.utils.circus import create_standalone_arbiter
    from .serving import API_SERVER_NAME
    from .serving import PROMETHEUS_MESSAGE
    from .serving import construct_ssl_args
    from .serving import construct_timeouts_args
    from .serving import create_watcher

    working_dir = os.path.realpath(os.path.expanduser(working_dir))
    svc = load(bento_identifier, working_dir=working_dir)
    timeout_args = ["--timeout", str(timeout)] if timeout else []
    runner_requirements = {runner.name for runner in svc.runners}
    if not runner_requirements.issubset(set(runner_map)):
        raise ValueError(
            f"{bento_identifier} requires runners {runner_requirements}, but only {set(runner_map)} are provided."
        )
    watchers: list[Watcher] = []
    circus_socket_map: dict[str, CircusSocket] = {}
    logger.debug("Runner map: %s", runner_map)
    circus_socket_map[API_SERVER_NAME] = CircusSocket(
        name=API_SERVER_NAME,
        host=host,
        port=port,
        backlog=backlog,
    )
    ssl_args = construct_ssl_args(
        ssl_certfile=ssl_certfile,
        ssl_keyfile=ssl_keyfile,
        ssl_keyfile_password=ssl_keyfile_password,
        ssl_version=ssl_version,
        ssl_cert_reqs=ssl_cert_reqs,
        ssl_ca_certs=ssl_ca_certs,
        ssl_ciphers=ssl_ciphers,
    )
    timeouts_args = construct_timeouts_args(
        timeout_keep_alive=timeout_keep_alive,
        timeout_graceful_shutdown=timeout_graceful_shutdown,
    )
    scheme = "https" if BentoMLContainer.ssl.enabled.get() else "http"
    watchers.append(
        create_watcher(
            name="api_server",
            args=[
                "-m",
                SCRIPT_API_SERVER,
                bento_identifier,
                "--fd",
                f"$(circus.sockets.{API_SERVER_NAME})",
                "--runner-map",
                json.dumps(runner_map),
                "--working-dir",
                working_dir,
                "--backlog",
                f"{backlog}",
                "--worker-id",
                "$(CIRCUS.WID)",
                *ssl_args,
                *timeouts_args,
                *timeout_args,
            ],
            working_dir=working_dir,
            numprocesses=api_workers,
            env=env,
        )
    )
    if BentoMLContainer.api_server_config.metrics.enabled.get():
        logger.info(
            PROMETHEUS_MESSAGE,
            scheme.upper(),
            bento_identifier,
            f"{scheme}://{host}:{port}/metrics",
        )

    arbiter = create_standalone_arbiter(
        watchers=watchers,
        sockets=list(circus_socket_map.values()),
    )
    with track_serve(svc, production=True, component=API_SERVER):
        arbiter.start(
            cb=lambda _: logger.info(  # type: ignore
                'Starting bare %s BentoServer from "%s" running on %s://%s:%d (Press CTRL+C to quit)',
                scheme.upper(),
                bento_identifier,
                scheme,
                host,
                port,
            ),
        )


@inject
def start_grpc_server(
    bento_identifier: str,
    runner_map: dict[str, str],
    working_dir: str,
    port: int = Provide[BentoMLContainer.grpc.port],
    host: str = Provide[BentoMLContainer.grpc.host],
    backlog: int = Provide[BentoMLContainer.api_server_config.backlog],
    api_workers: int = Provide[BentoMLContainer.api_server_workers],
    reflection: bool = Provide[BentoMLContainer.grpc.reflection.enabled],
    channelz: bool = Provide[BentoMLContainer.grpc.channelz.enabled],
    max_concurrent_streams: int | None = Provide[
        BentoMLContainer.grpc.max_concurrent_streams
    ],
    ssl_certfile: str | None = Provide[BentoMLContainer.ssl.certfile],
    ssl_keyfile: str | None = Provide[BentoMLContainer.ssl.keyfile],
    ssl_ca_certs: str | None = Provide[BentoMLContainer.ssl.ca_certs],
    protocol_version: str = LATEST_PROTOCOL_VERSION,
) -> None:
    from .serving import ensure_prometheus_dir

    env = {"PROMETHEUS_MULTIPROC_DIR": ensure_prometheus_dir()}

    from circus.sockets import CircusSocket
    from circus.watcher import Watcher

    from bentoml import load

    from ._internal.utils import reserve_free_port
    from ._internal.utils.analytics import track_serve
    from ._internal.utils.circus import create_standalone_arbiter
    from .serving import PROMETHEUS_MESSAGE
    from .serving import PROMETHEUS_SERVER_NAME
    from .serving import construct_ssl_args
    from .serving import create_watcher

    working_dir = os.path.realpath(os.path.expanduser(working_dir))
    svc = load(bento_identifier, working_dir=working_dir)
    runner_requirements = {runner.name for runner in svc.runners}
    if not runner_requirements.issubset(set(runner_map)):
        raise ValueError(
            f"{bento_identifier} requires runners {runner_requirements}, but only {set(runner_map)} are provided."
        )
    watchers: list[Watcher] = []
    circus_socket_map: dict[str, CircusSocket] = {}
    logger.debug("Runner map: %s", runner_map)
    ssl_args = construct_ssl_args(
        ssl_certfile=ssl_certfile,
        ssl_keyfile=ssl_keyfile,
        ssl_ca_certs=ssl_ca_certs,
    )
    scheme = "https" if BentoMLContainer.ssl.enabled.get() else "http"
    with contextlib.ExitStack() as port_stack:
        api_port = port_stack.enter_context(
            reserve_free_port(host=host, port=port, enable_so_reuseport=True)
        )

        args = [
            "-m",
            SCRIPT_GRPC_API_SERVER,
            bento_identifier,
            "--host",
            host,
            "--port",
            str(api_port),
            "--runner-map",
            json.dumps(runner_map),
            "--working-dir",
            working_dir,
            "--worker-id",
            "$(CIRCUS.WID)",
            *ssl_args,
            "--protocol-version",
            protocol_version,
        ]
        if reflection:
            args.append("--enable-reflection")
        if channelz:
            args.append("--enable-channelz")
        if max_concurrent_streams:
            args.extend(
                [
                    "--max-concurrent-streams",
                    str(max_concurrent_streams),
                ]
            )

        watchers.append(
            create_watcher(
                name="grpc_api_server",
                args=args,
                use_sockets=False,
                working_dir=working_dir,
                numprocesses=api_workers,
                env=env,
            )
        )

    if BentoMLContainer.api_server_config.metrics.enabled.get():
        metrics_host = BentoMLContainer.grpc.metrics.host.get()
        metrics_port = BentoMLContainer.grpc.metrics.port.get()

        circus_socket_map[PROMETHEUS_SERVER_NAME] = CircusSocket(
            name=PROMETHEUS_SERVER_NAME,
            host=metrics_host,
            port=metrics_port,
            backlog=backlog,
        )

        watchers.append(
            create_watcher(
                name="prom_server",
                args=[
                    "-m",
                    SCRIPT_GRPC_PROMETHEUS_SERVER,
                    "--fd",
                    f"$(circus.sockets.{PROMETHEUS_SERVER_NAME})",
                    "--backlog",
                    f"{backlog}",
                ],
                working_dir=working_dir,
                numprocesses=1,
                singleton=True,
                env=env,
            )
        )

        logger.info(
            PROMETHEUS_MESSAGE,
            "gRPC",
            bento_identifier,
            f"http://{metrics_host}:{metrics_port}",
        )
    arbiter = create_standalone_arbiter(
        watchers=watchers, sockets=list(circus_socket_map.values())
    )
    with track_serve(svc, production=True, component=API_SERVER, serve_kind="grpc"):
        arbiter.start(
            cb=lambda _: logger.info(  # type: ignore
                'Starting bare %s BentoServer from "%s" running on %s://%s:%d (Press CTRL+C to quit)',
                "gRPC",
                bento_identifier,
                scheme,
                host,
                port,
            ),
        )



from __future__ import annotations

import inspect
import io
import json
import logging
import mimetypes
import os
import pathlib
import tempfile
import time
import typing as t
from abc import abstractmethod
from functools import cached_property
from http import HTTPStatus
from urllib.parse import urljoin
from urllib.parse import urlparse

import attr
import httpx

from _bentoml_sdk import IODescriptor
from _bentoml_sdk.typing_utils import is_image_type
from bentoml import __version__
from bentoml._internal.utils.uri import is_http_url
from bentoml._internal.utils.uri import uri_to_path
from bentoml.exceptions import BentoMLException
from bentoml.exceptions import NotFound
from bentoml.exceptions import ServiceUnavailable

from ..serde import Payload
from ..tasks import ResultStatus
from .base import AbstractClient
from .base import ClientEndpoint
from .base import map_exception
from .task import AsyncTask
from .task import Task

if t.TYPE_CHECKING:
    from httpx._types import RequestFiles
    from PIL import Image

    from _bentoml_sdk import Service
    from bentoml._internal.external_typing import ASGIApp

    from ..serde import Serde

    T = t.TypeVar("T", bound="HTTPClient[t.Any]")
    A = t.TypeVar("A")
C = t.TypeVar("C", httpx.Client, httpx.AsyncClient)
AnyClient = t.TypeVar("AnyClient", httpx.Client, httpx.AsyncClient)
logger = logging.getLogger("bentoml.io")
MAX_RETRIES = 3


def to_async_iterable(iterable: t.Iterable[A]) -> t.AsyncIterable[A]:
    async def _gen() -> t.AsyncIterator[A]:
        for item in iterable:
            yield item

    return _gen()


@attr.define
class HTTPClient(AbstractClient, t.Generic[C]):
    client_cls: t.ClassVar[type[httpx.Client] | type[httpx.AsyncClient]]

    url: str
    endpoints: dict[str, ClientEndpoint] = attr.field(factory=dict)
    media_type: str = "application/json"
    timeout: float = 30
    default_headers: dict[str, str] = attr.field(factory=dict)
    app: ASGIApp | None = None

    _opened_files: list[io.BufferedReader] = attr.field(init=False, factory=list)
    _temp_dir: tempfile.TemporaryDirectory[str] = attr.field(init=False)

    @staticmethod
    def _make_client(
        client_cls: type[AnyClient],
        url: str,
        headers: t.Mapping[str, str],
        timeout: float,
        app: ASGIApp | None = None,
    ) -> AnyClient:
        parsed = urlparse(url)
        transport = None
        if parsed.scheme == "file":
            uds = uri_to_path(url)
            if client_cls is httpx.Client:
                transport = httpx.HTTPTransport(uds=uds)
            else:
                transport = httpx.AsyncHTTPTransport(uds=uds)
            url = "http://127.0.0.1:3000"
        elif parsed.scheme == "tcp":
            url = f"http://{parsed.netloc}"
        elif app is not None:
            if client_cls is httpx.Client:
                from a2wsgi import ASGIMiddleware

                transport = httpx.WSGITransport(app=ASGIMiddleware(app))
            else:
                transport = httpx.ASGITransport(app=app)
        return client_cls(
            base_url=url,
            transport=transport,  # type: ignore
            headers=headers,
            timeout=timeout,
            follow_redirects=True,
        )

    @_temp_dir.default  # type: ignore
    def default_temp_dir(self) -> tempfile.TemporaryDirectory[str]:
        return tempfile.TemporaryDirectory(prefix="bentoml-client-")

    def __init__(
        self,
        url: str,
        *,
        media_type: str = "application/json",
        service: Service[t.Any] | None = None,
        server_ready_timeout: float | None = None,
        token: str | None = None,
        timeout: float = 30,
        app: ASGIApp | None = None,
    ) -> None:
        """Create a client instance from a URL.

        Args:
            url: The URL of the BentoML service.
            media_type: The media type to use for serialization. Defaults to
                "application/json".

        .. note::

            The client created with this method can only return primitive types without a model.
        """
        routes: dict[str, ClientEndpoint] = {}
        default_headers = {"User-Agent": f"BentoML HTTP Client/{__version__}"}
        if token is None:
            token = os.getenv("BENTO_CLOUD_API_KEY")
        if token:
            default_headers["Authorization"] = f"Bearer {token}"

        if service is not None:
            for name, method in service.apis.items():
                routes[name] = ClientEndpoint(
                    name=name,
                    route=method.route,
                    input=method.input_spec.model_json_schema(),
                    output=method.output_spec.model_json_schema(),
                    doc=method.doc,
                    input_spec=method.input_spec,
                    output_spec=method.output_spec,
                    stream_output=method.is_stream,
                    is_task=method.is_task,
                )

            from bentoml._internal.context import server_context

            default_headers.update(
                {
                    "Bento-Name": server_context.bento_name,
                    "Bento-Version": server_context.bento_version,
                    "Runner-Name": service.name,
                    "Yatai-Bento-Deployment-Name": server_context.yatai_bento_deployment_name,
                    "Yatai-Bento-Deployment-Namespace": server_context.yatai_bento_deployment_namespace,
                }
            )
        self.__attrs_init__(  # type: ignore
            url=url,
            endpoints=routes,
            media_type=media_type,
            default_headers=default_headers,
            timeout=timeout,
            app=app,
        )
        if app is None and (server_ready_timeout is None or server_ready_timeout > 0):
            self.wait_until_server_ready(server_ready_timeout)
        if service is None:
            schema_url = urljoin(url, "/schema.json")

            with self._make_client(
                httpx.Client, url, default_headers, timeout, app=app
            ) as client:
                resp = client.get("/schema.json")

                if resp.is_error:
                    raise BentoMLException(f"Failed to fetch schema from {schema_url}")
                for route in resp.json()["routes"]:
                    self.endpoints[route["name"]] = ClientEndpoint(
                        name=route["name"],
                        route=route["route"],
                        input=route["input"],
                        output=route["output"],
                        doc=route.get("doc"),
                        stream_output=route["output"].get("is_stream", False),
                        is_task=route.get("is_task", False),
                    )
        super().__init__()

    @cached_property
    def client(self) -> C:
        return self._make_client(
            self.client_cls, self.url, self.default_headers, self.timeout, self.app
        )

    @cached_property
    def serde(self) -> Serde:
        from ..serde import ALL_SERDE

        return ALL_SERDE[self.media_type]()

    def _build_request(
        self,
        endpoint: ClientEndpoint,
        args: t.Sequence[t.Any],
        kwargs: dict[str, t.Any],
        headers: t.Mapping[str, str],
    ) -> httpx.Request:
        from opentelemetry import propagate

        from _bentoml_sdk.io_models import IORootModel

        headers = httpx.Headers({"Content-Type": self.media_type, **headers})
        propagate.inject(headers)
        if endpoint.input_spec is not None:
            model = endpoint.input_spec.from_inputs(*args, **kwargs)
            if (
                not isinstance(model, IORootModel)
                and model.multipart_fields
                and self.media_type == "application/json"
            ):
                return self._build_multipart(endpoint, model, headers)
            elif isinstance(rendered := model.model_dump(), (str, bytes)):
                headers.update({"content-type": model.mime_type()})
                return self.client.build_request(
                    "POST", endpoint.route, content=rendered, headers=headers
                )
            else:
                payload = self.serde.serialize_model(model)
                headers.update(payload.headers)
                return self.client.build_request(
                    "POST",
                    endpoint.route,
                    headers=headers,
                    content=to_async_iterable(payload.data)
                    if self.client_cls is httpx.AsyncClient
                    else payload.data,
                )
        assert self.media_type == "application/json", (
            "Non-JSON request is not supported"
        )
        if endpoint.input.get("root_input", False):
            if len(args) > 1 or kwargs:
                raise TypeError("Expected one positional argument for root input")
            if not args:
                return self.client.build_request(
                    "POST", endpoint.route, headers=headers
                )
            value = args[0]
            passthrough = False
            content = None
            if "properties" in endpoint.input:
                kwargs = value
                args = ()
                passthrough = True
            elif endpoint.input.get("type") == "file":
                file = self._get_file(value)
                if isinstance(file, str):
                    content = file
                else:
                    file_io, content_type = file[1:]
                    content = iter(lambda: file_io.read(4096), b"")
                    if content_type:
                        headers.update({"content-type": content_type})
            elif isinstance(value, (str, bytes)):
                content = value.encode("utf-8") if isinstance(value, str) else value
                headers.update({"content-type": "text/plain"})
            else:
                payload = self.serde.serialize(value, endpoint.input)
                headers.update(payload.headers)
                content = (
                    to_async_iterable(payload.data)
                    if self.client_cls is httpx.AsyncClient
                    else payload.data
                )
            if not passthrough:
                return self.client.build_request(
                    "POST", endpoint.route, content=content, headers=headers
                )

        for name, value in zip(endpoint.input["properties"], args):
            if name in kwargs:
                raise TypeError(f"Duplicate argument {name}")
            kwargs[name] = value

        params = set(endpoint.input["properties"].keys())
        non_exist_args = set(kwargs.keys()) - set(params)
        if non_exist_args:
            raise TypeError(
                f"Arguments not found in endpoint {endpoint.name}: {non_exist_args}"
            )
        required = set(endpoint.input.get("required", []))
        missing_args = set(required) - set(kwargs.keys())
        if missing_args:
            raise TypeError(
                f"Missing required arguments in endpoint {endpoint.name}: {missing_args}"
            )
        has_file = any(
            schema.get("type") == "file"
            or schema.get("type") == "array"
            and schema["items"].get("type") == "file"
            for schema in endpoint.input["properties"].values()
        )
        if has_file:
            return self._build_multipart(endpoint, kwargs, headers)
        payload = self.serde.serialize(kwargs, endpoint.input)
        headers.update(payload.headers)
        return self.client.build_request(
            "POST",
            endpoint.route,
            content=to_async_iterable(payload.data)
            if self.client_cls is httpx.AsyncClient
            else payload.data,
            headers=headers,
        )

    def wait_until_server_ready(self, timeout: int | None = None) -> None:
        if timeout is None:
            timeout = self.timeout
        with self._make_client(
            httpx.Client, self.url, self.default_headers, timeout
        ) as client:
            start = time.monotonic()
            while time.monotonic() - start < timeout:
                try:
                    resp = client.get("/readyz")
                    if resp.status_code == 200:
                        return
                except (httpx.TimeoutException, httpx.ConnectError):
                    pass
        raise ServiceUnavailable(f"Server is not ready after {timeout} seconds")

    def _get_file(self, value: t.Any) -> str | tuple[str, t.IO[bytes], str | None]:
        if isinstance(value, str) and not is_http_url(value):
            value = pathlib.Path(value)
        if is_image_type(type(value)):
            fp = getattr(value, "_fp", value.fp)
            fname = getattr(fp, "name", None)
            fmt = value.format.lower()
            return (
                pathlib.Path(fname).name if fname else f"upload-image.{fmt}",
                fp,
                f"image/{fmt}",
            )
        elif isinstance(value, pathlib.PurePath):
            file = open(value, "rb")
            self._opened_files.append(file)
            return (value.name, file, mimetypes.guess_type(value)[0])
        elif isinstance(value, str):
            return value
        else:
            assert isinstance(value, t.BinaryIO)
            filename = pathlib.Path(getattr(value, "name", "upload-file")).name
            content_type = mimetypes.guess_type(filename)[0]
            return (filename, value, content_type)

    def _build_multipart(
        self,
        endpoint: ClientEndpoint,
        model: IODescriptor | dict[str, t.Any],
        headers: httpx.Headers,
    ) -> httpx.Request:
        def is_file_field(k: str) -> bool:
            if isinstance(model, IODescriptor):
                return k in model.multipart_fields
            if (f := endpoint.input["properties"].get(k, {})).get("type") == "file":
                return True
            if f.get("type") == "array" and f["items"].get("type") == "file":
                return True
            return False

        if isinstance(model, dict):
            fields = model
        else:
            fields = {k: getattr(model, k) for k in model.model_fields}
        data: dict[str, t.Any] = {}
        files: RequestFiles = []

        for name, value in fields.items():
            if not is_file_field(name):
                data[name] = json.dumps(value)
                continue
            if not isinstance(value, (list, tuple)):
                value = [value]

            for v in value:
                file = self._get_file(v)
                if isinstance(file, str):
                    data[name] = file
                else:
                    files.append((name, file))
        headers.pop("content-type", None)
        return self.client.build_request(
            "POST", endpoint.route, data=data, files=files, headers=headers
        )

    def _deserialize_output(self, payload: Payload, endpoint: ClientEndpoint) -> t.Any:
        from _bentoml_sdk.io_models import IORootModel

        data = iter(payload.data)
        if (endpoint.output.get("type")) == "string":
            content = bytes(next(data))
            if endpoint.output.get("format") == "binary":
                return content
            return content.decode("utf-8")
        elif endpoint.output_spec is not None:
            model = self.serde.deserialize_model(payload, endpoint.output_spec)
            if isinstance(model, IORootModel):
                return model.root  # type: ignore
            return model
        else:
            return self.serde.deserialize(payload, endpoint.output)

    def call(self, __name: str, /, *args: t.Any, **kwargs: t.Any) -> t.Any:
        try:
            endpoint = self.endpoints[__name]
        except KeyError:
            raise NotFound(f"Endpoint {__name} not found") from None
        if endpoint.stream_output:
            return self._get_stream(endpoint, args, kwargs)
        else:
            return self._call(endpoint, args, kwargs)

    @abstractmethod
    def _call(
        self,
        endpoint: ClientEndpoint,
        args: t.Sequence[t.Any],
        kwargs: dict[str, t.Any],
        *,
        headers: t.Mapping[str, str] | None = None,
    ) -> t.Any: ...

    @abstractmethod
    def _get_stream(
        self, endpoint: ClientEndpoint, args: t.Any, kwargs: t.Any
    ) -> t.Any: ...


class SyncHTTPClient(HTTPClient[httpx.Client]):
    """A synchronous client for BentoML service.

    .. note:: Inner usage ONLY
    """

    client_cls = httpx.Client

    def __enter__(self: T) -> T:
        return self

    def __exit__(self, exc_type: t.Any, exc: t.Any, tb: t.Any) -> None:
        return self.close()

    def is_ready(self, timeout: int | None = None) -> bool:
        try:
            resp = self.client.get(
                "/readyz", timeout=timeout or httpx.USE_CLIENT_DEFAULT
            )
            return resp.status_code == 200
        except httpx.TimeoutException:
            logger.warn("Timed out waiting for runner to be ready")
            return False

    def close(self) -> None:
        if "client" in vars(self):
            self.client.close()

    def _get_stream(
        self, endpoint: ClientEndpoint, args: t.Any, kwargs: t.Any
    ) -> t.Generator[t.Any, None, None]:
        resp = self._call(endpoint, args, kwargs)
        for data in resp:
            yield data

    def request(self, method: str, url: str, **kwargs: t.Any) -> httpx.Response:
        return self.client.request(method, url, **kwargs)

    def _submit(
        self, __endpoint: ClientEndpoint, /, *args: t.Any, **kwargs: t.Any
    ) -> Task:
        try:
            req = self._build_request(__endpoint, args, kwargs, {})
            req.url = req.url.copy_with(path=f"{__endpoint.route}/submit")
            resp = self.client.send(req)
            if resp.is_error:
                resp.read()
                raise BentoMLException(
                    f"Error making request: {resp.status_code}: {resp.text}",
                    error_code=HTTPStatus(resp.status_code),
                )
            data = resp.json()
            return Task(data["task_id"], __endpoint, self)
        finally:
            for f in self._opened_files:
                f.close()
            self._opened_files.clear()

    def _get_task_result(self, __endpoint: ClientEndpoint, /, task_id: str) -> t.Any:
        resp = self.request("GET", f"{__endpoint}/get", params={"task_id": task_id})
        if resp.is_error:
            resp.read()
            raise map_exception(resp)
        if (
            __endpoint.output.get("type") == "file"
            and self.media_type == "application/json"
        ):
            return self._parse_file_response(__endpoint, resp)
        else:
            return self._parse_response(__endpoint, resp)

    def _get_task_status(
        self, __endpoint: ClientEndpoint, /, task_id: str
    ) -> ResultStatus:
        resp = self.client.request(
            "GET", f"{__endpoint.route}/status", params={"task_id": task_id}
        )
        if resp.is_error:
            resp.read()
            raise map_exception(resp)
        data = resp.json()
        return ResultStatus(data["status"])

    def _cancel_task(self, __endpoint: ClientEndpoint, /, task_id: str) -> None:
        resp = self.request(
            "PUT", f"{__endpoint.route}/cancel", params={"task_id": task_id}
        )
        if resp.is_error:
            resp.read()
            raise map_exception(resp)

    def _retry_task(self, __endpoint: ClientEndpoint, /, task_id: str) -> Task:
        resp = self.request(
            "POST", f"{__endpoint.route}/retry", params={"task_id": task_id}
        )
        if resp.is_error:
            resp.read()
            raise map_exception(resp)
        data = resp.json()
        return Task(data["task_id"], __endpoint, self)

    def _call(
        self,
        endpoint: ClientEndpoint,
        args: t.Sequence[t.Any],
        kwargs: dict[str, t.Any],
        *,
        headers: t.Mapping[str, str] | None = None,
    ) -> t.Any:
        try:
            req = self._build_request(endpoint, args, kwargs, headers or {})
            resp = self.client.send(req, stream=endpoint.stream_output)
            if resp.is_error:
                resp.read()
                raise map_exception(resp)
            if endpoint.stream_output:
                return self._parse_stream_response(endpoint, resp)
            elif endpoint.output.get("type") == "file":
                # file responses are always raw binaries whatever the serde is
                return self._parse_file_response(endpoint, resp)
            else:
                return self._parse_response(endpoint, resp)
        finally:
            for f in self._opened_files:
                f.close()
            self._opened_files.clear()

    def _parse_response(self, endpoint: ClientEndpoint, resp: httpx.Response) -> t.Any:
        payload = Payload((resp.read(),), resp.headers)
        return self._deserialize_output(payload, endpoint)

    def _parse_stream_response(
        self, endpoint: ClientEndpoint, resp: httpx.Response
    ) -> t.Generator[t.Any, None, None]:
        try:
            for data in resp.iter_bytes():
                yield self._deserialize_output(Payload((data,), resp.headers), endpoint)
        finally:
            resp.close()

    def _parse_file_response(
        self, endpoint: ClientEndpoint, resp: httpx.Response
    ) -> pathlib.Path | Image.Image:
        from multipart.multipart import parse_options_header
        from PIL import Image

        content_disposition = resp.headers.get("content-disposition")
        content_type = resp.headers.get("content-type", "")
        filename: str | None = None
        if endpoint.output.get("pil"):
            image_formats = (
                [content_type[6:]] if content_type.startswith("image/") else None
            )
            return Image.open(io.BytesIO(resp.read()), formats=image_formats)
        if content_disposition:
            _, options = parse_options_header(content_disposition)
            if b"filename" in options:
                filename = str(
                    options[b"filename"],
                    resp.charset_encoding or "utf-8",
                    errors="ignore",
                )

        with tempfile.NamedTemporaryFile(
            "wb", suffix=filename, dir=self._temp_dir.name, delete=False
        ) as f:
            f.write(resp.read())
        return pathlib.Path(f.name)


class AsyncHTTPClient(HTTPClient[httpx.AsyncClient]):
    """An asynchronous client for BentoML service.

    .. note:: Inner usage ONLY
    """

    client_cls = httpx.AsyncClient

    async def is_ready(self, timeout: int | None = None) -> bool:
        try:
            resp = await self.client.get(
                "/readyz", timeout=timeout or httpx.USE_CLIENT_DEFAULT
            )
            return resp.status_code == 200
        except httpx.TimeoutException:
            logger.warn("Timed out waiting for runner to be ready")
            return False

    async def _get_stream(
        self, endpoint: ClientEndpoint, args: t.Any, kwargs: t.Any
    ) -> t.AsyncGenerator[t.Any, None]:
        resp = await self._call(endpoint, args, kwargs)
        assert inspect.isasyncgen(resp)
        async for data in resp:
            yield data

    async def __aenter__(self: T) -> T:
        return self

    async def __aexit__(self, *args: t.Any) -> None:
        return await self.close()

    async def request(self, method: str, url: str, **kwargs: t.Any) -> httpx.Response:
        return await self.client.request(method, url, **kwargs)

    async def _submit(
        self, __endpoint: ClientEndpoint, /, *args: t.Any, **kwargs: t.Any
    ) -> AsyncTask:
        try:
            req = self._build_request(__endpoint, args, kwargs, {})
            req.url = req.url.copy_with(path=f"{__endpoint.route}/submit")
            resp = await self.client.send(req)
            if resp.is_error:
                resp.read()
                raise BentoMLException(
                    f"Error making request: {resp.status_code}: {resp.text}",
                    error_code=HTTPStatus(resp.status_code),
                )
            data = resp.json()
            return AsyncTask(data["task_id"], __endpoint, self)
        finally:
            for f in self._opened_files:
                f.close()
            self._opened_files.clear()

    async def _get_task_status(
        self, __endpoint: ClientEndpoint, /, task_id: str
    ) -> ResultStatus:
        resp = await self.client.request(
            "GET", f"{__endpoint.route}/status", params={"task_id": task_id}
        )
        if resp.is_error:
            await resp.aread()
            raise map_exception(resp)
        data = resp.json()
        return ResultStatus(data["status"])

    async def _cancel_task(self, __endpoint: ClientEndpoint, /, task_id: str) -> None:
        resp = await self.request(
            "PUT", f"{__endpoint.route}/cancel", params={"task_id": task_id}
        )
        if resp.is_error:
            await resp.aread()
            raise map_exception(resp)

    async def _retry_task(
        self, __endpoint: ClientEndpoint, /, task_id: str
    ) -> AsyncTask:
        resp = await self.request(
            "POST", f"{__endpoint.route}/retry", params={"task_id": task_id}
        )
        if resp.is_error:
            await resp.aread()
            raise map_exception(resp)
        data = resp.json()
        return AsyncTask(data["task_id"], __endpoint, self)

    async def _get_task_result(
        self, __endpoint: ClientEndpoint, /, task_id: str
    ) -> t.Any:
        resp = await self.request(
            "GET", f"{__endpoint.route}/get", params={"task_id": task_id}
        )
        if resp.is_error:
            await resp.aread()
            raise map_exception(resp)
        if (
            __endpoint.output.get("type") == "file"
            and self.media_type == "application/json"
        ):
            return await self._parse_file_response(__endpoint, resp)
        else:
            return await self._parse_response(__endpoint, resp)

    async def _call(
        self,
        endpoint: ClientEndpoint,
        args: t.Sequence[t.Any],
        kwargs: dict[str, t.Any],
        *,
        headers: t.Mapping[str, str] | None = None,
    ) -> t.Any:
        try:
            req = self._build_request(endpoint, args, kwargs, headers or {})
            resp = await self.client.send(req, stream=endpoint.stream_output)
            if resp.is_error:
                await resp.aread()
                raise map_exception(resp)
            if endpoint.stream_output:
                return self._parse_stream_response(endpoint, resp)
            elif endpoint.output.get("type") == "file":
                # file responses are always raw binaries whatever the serde is
                return await self._parse_file_response(endpoint, resp)
            else:
                return await self._parse_response(endpoint, resp)
        finally:
            for f in self._opened_files:
                f.close()
            self._opened_files.clear()

    async def _parse_response(
        self, endpoint: ClientEndpoint, resp: httpx.Response
    ) -> t.Any:
        data = await resp.aread()
        return self._deserialize_output(Payload((data,), resp.headers), endpoint)

    async def _parse_stream_response(
        self, endpoint: ClientEndpoint, resp: httpx.Response
    ) -> t.AsyncGenerator[t.Any, None]:
        try:
            async for data in resp.aiter_bytes():
                yield self._deserialize_output(Payload((data,), resp.headers), endpoint)
        finally:
            await resp.aclose()

    async def _parse_file_response(
        self, endpoint: ClientEndpoint, resp: httpx.Response
    ) -> pathlib.Path | Image.Image:
        from multipart.multipart import parse_options_header
        from PIL import Image

        content_disposition = resp.headers.get("content-disposition")
        content_type = resp.headers.get("content-type", "")
        filename: str | None = None
        if endpoint.output.get("pil"):
            image_formats = (
                [content_type[6:]] if content_type.startswith("image/") else None
            )
            return Image.open(io.BytesIO(await resp.aread()), formats=image_formats)
        if content_disposition:
            _, options = parse_options_header(content_disposition)
            if b"filename" in options:
                filename = str(
                    options[b"filename"],
                    resp.charset_encoding or "utf-8",
                    errors="ignore",
                )

        with tempfile.NamedTemporaryFile(
            "wb", suffix=filename, dir=self._temp_dir.name, delete=False
        ) as f:
            f.write(await resp.aread())
        return pathlib.Path(f.name)

    async def close(self) -> None:
        if "client" in vars(self):
            await self.client.aclose()



from __future__ import annotations

import abc
import asyncio
import base64
import io
import json
import typing as t

from starlette.requests import Request
from starlette.responses import Response

if t.TYPE_CHECKING:
    from starlette.types import Message


async def _consume_response(response: Response) -> bytes:
    # This will collect the response body into a buffer
    # and run until the background task is done
    buffer = io.BytesIO()
    scope: dict[str, t.Any] = {"type": "http"}

    def receive() -> t.Awaitable[Message]:
        # A fake receive that never returns
        loop = asyncio.get_running_loop()
        fut = loop.create_future()
        return fut

    async def send(message: Message) -> None:
        if message.get("type") != "http.response.body":
            return
        buffer.write(message.get("body", b""))

    await response(scope, receive, send)
    return buffer.getvalue()


class Serde(abc.ABC):
    @abc.abstractmethod
    async def serialize_request(self, request: Request) -> bytes:
        raise NotImplementedError

    @abc.abstractmethod
    async def deserialize_request(self, data: bytes) -> Request:
        raise NotImplementedError

    @abc.abstractmethod
    async def serialize_response(self, response: Response) -> bytes:
        raise NotImplementedError

    @abc.abstractmethod
    async def deserialize_response(self, data: bytes) -> Response:
        raise NotImplementedError


class JSONSerde(Serde):
    HEADERS_ENCODING = "iso-8859-1"

    def get_scope(self, scope: dict[str, t.Any]) -> dict[str, t.Any]:
        encodable_fields = [
            "type",
            "asgi",
            "http_version",
            "method",
            "scheme",
            "path",
            "root_path",
            "raw_path",
            "query_string",
            "headers",
        ]
        new_scope = {k: v for k, v in scope.items() if k in encodable_fields}
        return new_scope

    def json_encode(self, obj: t.Any) -> bytes:
        def default(obj: t.Any) -> t.Any:
            if isinstance(obj, bytes):
                return obj.decode(self.HEADERS_ENCODING)
            raise TypeError(
                f"Object of type {obj.__class__.__name__} is not JSON serializable"
            )

        return json.dumps(obj, separators=(",", ":"), default=default).encode()

    async def serialize_request(self, request: Request) -> bytes:
        request_dict = {
            "scope": self.get_scope(request.scope),
            "content": base64.b64encode(await request.body()).decode("ascii"),
        }
        return self.json_encode(request_dict)

    async def deserialize_request(self, data: bytes) -> Request:
        request_dict = json.loads(data)
        scope = request_dict["scope"]
        for field in ("raw_path", "query_string"):
            if field in scope:
                scope[field] = scope[field].encode(self.HEADERS_ENCODING)
        if "headers" in scope:
            scope["headers"] = [
                (k.encode(self.HEADERS_ENCODING), v.encode(self.HEADERS_ENCODING))
                for k, v in scope["headers"]
            ]
        request = Request(scope)
        request._body = base64.b64decode(request_dict["content"].encode("ascii"))
        return request

    async def serialize_response(self, response: Response) -> bytes:
        response_dict = {
            "status": response.status_code,
            "headers": list(response.headers.items()),
            "content": base64.b64encode(await _consume_response(response)).decode(
                "ascii"
            ),
        }
        return json.dumps(response_dict, separators=(",", ":")).encode()

    async def deserialize_response(self, data: bytes) -> Response:
        response_dict = json.loads(data)
        response = Response(
            content=base64.b64decode(response_dict["content"].encode("ascii")),
            status_code=response_dict["status"],
        )
        response.raw_headers = [
            tuple(map(str.encode, h)) for h in response_dict["headers"]
        ]
        return response



from __future__ import annotations

import logging
from contextvars import ContextVar
from timeit import default_timer
from typing import TYPE_CHECKING
from typing import Sequence

if TYPE_CHECKING:
    from ... import external_typing as ext

REQ_CONTENT_LENGTH = "REQUEST_CONTENT_LENGTH"
REQ_CONTENT_TYPE = "REQUEST_CONTENT_TYPE"
RESP_CONTENT_LENGTH = "RESPONSE_CONTENT_LENGTH"
RESP_CONTENT_TYPE = "RESPONSE_CONTENT_TYPE"

CONTENT_LENGTH = b"content-length"
CONTENT_TYPE = b"content-type"

status: ContextVar[int] = ContextVar("ACCESS_LOG_STATUS_CODE")
request_content_length: ContextVar[bytes] = ContextVar(
    "ACCESS_LOG_REQ_CONTENT_LENGTH",
    default=b"",
)
request_content_type: ContextVar[bytes] = ContextVar(
    "ACCESS_LOG_REQ_CONTENT_TYPE",
    default=b"",
)
response_content_length: ContextVar[bytes] = ContextVar(
    "ACCESS_LOG_RESP_CONTENT_LENGTH",
    default=b"",
)
response_content_type: ContextVar[bytes] = ContextVar(
    "ACCESS_LOG_RESP_CONTENT_TYPE",
    default=b"",
)


class AccessLogMiddleware:
    """
    ASGI Middleware implementation that intercepts and decorates the send
    and receive callables to generate the BentoML access log.
    """

    def __init__(
        self,
        app: ext.ASGIApp,
        has_request_content_length: bool = False,
        has_request_content_type: bool = False,
        has_response_content_length: bool = False,
        has_response_content_type: bool = False,
        skip_paths: Sequence[str] = (),
    ) -> None:
        self.app = app
        self.has_request_content_length = has_request_content_length
        self.has_request_content_type = has_request_content_type
        self.has_response_content_length = has_response_content_length
        self.has_response_content_type = has_response_content_type
        self.logger = logging.getLogger("bentoml.access")
        self.skip_paths = skip_paths

    async def __call__(
        self,
        scope: ext.ASGIScope,
        receive: ext.ASGIReceive,
        send: ext.ASGISend,
    ) -> None:
        if not scope["type"].startswith(("http", "websocket")):
            await self.app(scope, receive, send)
            return

        start = default_timer()
        client = scope.get("client")
        scheme = scope["scheme"]
        method = scope.get("method", "")
        path = scope["path"]
        if path.startswith(tuple(self.skip_paths)):
            await self.app(scope, receive, send)
            return

        if client:
            address = f"{client[0]}:{client[1]}"
        else:
            address = "_"

        if self.has_request_content_length or self.has_request_content_type:
            for key, value in scope["headers"]:
                if key == CONTENT_LENGTH:
                    request_content_length.set(value)
                elif key == CONTENT_TYPE:
                    request_content_type.set(value)

        async def wrapped_receive() -> "ext.ASGIMessage":
            message = await receive()
            if message["type"] == "websocket.connect":
                self.logger.info(
                    "%s (scheme=%s,path=%s) - Client connected", address, scheme, path
                )
            elif message["type"] == "websocket.disconnect":
                self.logger.info(
                    "%s (scheme=%s,path=%s) - Client disconnected",
                    address,
                    scheme,
                    path,
                )
            return message

        async def wrapped_send(message: "ext.ASGIMessage") -> None:
            latency = max(default_timer() - start, 0) * 1000
            if message["type"] == "http.response.start":
                status.set(message["status"])
                if self.has_response_content_length or self.has_response_content_type:
                    for key, value in message["headers"]:
                        if key == CONTENT_LENGTH:
                            response_content_length.set(value)
                        elif key == CONTENT_TYPE:
                            response_content_type.set(value)
            elif message["type"] == "websocket.close":
                self.logger.info(
                    "%s (scheme=%s,path=%s) - Connection closed", address, scheme, path
                )
            elif message["type"] == "websocket.accept":
                self.logger.info(
                    "%s (scheme=%s,path=%s) - Connection established",
                    address,
                    scheme,
                    path,
                )

            elif message["type"] == "http.response.body":
                if "more_body" in message and message["more_body"]:
                    await send(message)
                    return

                request = [f"scheme={scheme}", f"method={method}", f"path={path}"]
                if self.has_request_content_type:
                    request.append(f"type={request_content_type.get().decode()}")
                if self.has_request_content_length:
                    request.append(f"length={request_content_length.get().decode()}")

                response = [f"status={status.get()}"]
                if self.has_response_content_type:
                    response.append(f"type={response_content_type.get().decode()}")
                if self.has_response_content_length:
                    response.append(f"length={response_content_length.get().decode()}")

                await send(message)
                self.logger.info(
                    "%s (%s) (%s) %.3fms",
                    address,
                    ",".join(request),
                    ",".join(response),
                    latency,
                )
                return

            await send(message)

        await self.app(scope, wrapped_receive, wrapped_send)
